"""
The CollectionInfo object definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Dict, Optional

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    StringListDescriptor, SerializableDescriptor, ParametersDescriptor
from .base import DEFAULT_STRICT


class RadarModeType(Serializable):
    """
    Radar mode type container class
    """

    _fields = ('ModeType', 'ModeID')
    _required = ('ModeType', )
    # other class variable
    _MODE_TYPE_VALUES = ('SPOTLIGHT', 'STRIPMAP', 'DYNAMIC STRIPMAP')
    # descriptors
    ModeType = StringEnumDescriptor(
        'ModeType', _MODE_TYPE_VALUES, _required, strict=True,
        docstring="The Radar imaging mode.")  # type: str
    ModeID = StringDescriptor(
        'ModeID', _required, strict=DEFAULT_STRICT,
        docstring='Radar imaging mode per Program Specific Implementation Document.')  # type: str

    def __init__(
            self,
            ModeType: str = None,
            ModeID: Optional[str] = None,
            **kwargs):
        """

        Parameters
        ----------
        ModeType : str
        ModeID : None|str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ModeID = ModeID
        self.ModeType = ModeType
        super(RadarModeType, self).__init__(**kwargs)

    def get_mode_abbreviation(self) -> str:
        """
        Get the mode abbreviation for the suggested name.

        Returns
        -------
        str
        """

        mode = self.ModeType
        if mode == 'SPOTLIGHT':
            return 'SL'
        elif mode == 'STRIPMAP':
            return 'ST'
        elif mode == 'DYNAMIC STRIPMAP':
            return 'DS'


class CollectionInfoType(Serializable):
    """General information about the collection."""
    _collections_tags = {
        'Parameters': {'array': False, 'child_tag': 'Parameter'},
        'CountryCodes': {'array': False, 'child_tag': 'CountryCode'},
    }
    _fields = (
        'CollectorName', 'IlluminatorName', 'CoreName', 'CollectType',
        'RadarMode', 'Classification', 'CountryCodes', 'Parameters')
    _required = ('CollectorName', 'CoreName', 'RadarMode', 'Classification')
    # other class variable
    _COLLECT_TYPE_VALUES = ('MONOSTATIC', 'BISTATIC')
    # descriptors
    CollectorName = StringDescriptor(
        'CollectorName', _required, strict=DEFAULT_STRICT,
        docstring='Radar platform identifier. For Bistatic collections, list the Receive platform.')  # type: str
    IlluminatorName = StringDescriptor(
        'IlluminatorName', _required, strict=DEFAULT_STRICT,
        docstring='Radar platform identifier that provided the illumination. For Bistatic collections, '
                  'list the transmit platform.')  # type: str
    CoreName = StringDescriptor(
        'CoreName', _required, strict=DEFAULT_STRICT,
        docstring='Collection and imaging data set identifier. Uniquely identifies imaging collections per '
                  'Program Specific Implementation Doc.')  # type: str
    CollectType = StringEnumDescriptor(
        'CollectType', _COLLECT_TYPE_VALUES, _required,
        docstring="Collection type identifier. Monostatic collections include single platform collections with "
                  "unique transmit and receive apertures.")  # type: str
    RadarMode = SerializableDescriptor(
        'RadarMode', RadarModeType, _required, strict=DEFAULT_STRICT,
        docstring='The radar mode.')  # type: RadarModeType
    Classification = StringDescriptor(
        'Classification', _required, strict=DEFAULT_STRICT, default_value='UNCLASSIFIED',
        docstring='Contains the human-readable banner. Contains classification, file control and handling, '
                  'file releasing, and/or proprietary markings. Specified per Program Specific '
                  'Implementation Document.')  # type: str
    CountryCodes = StringListDescriptor(
        'CountryCodes', _required, strict=DEFAULT_STRICT,
        docstring="List of country codes for region covered by the image.")  # type: List[str]
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Free form parameters object collection.')  # type: ParametersCollection

    def __init__(
            self,
            CollectorName: str = None,
            IlluminatorName: Optional[str] = None,
            CoreName: str = None,
            CollectType: Optional[str] = None,
            RadarMode: RadarModeType = None,
            Classification: str = "UNCLASSIFIED",
            CountryCodes: Union[str, List[str]] = None,
            Parameters: Union[ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        CollectorName : str
        IlluminatorName : str
        CoreName : str
        CollectType : str
        RadarMode : RadarModeType
        Classification : str
        CountryCodes : list|str
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectorName = CollectorName
        self.IlluminatorName = IlluminatorName
        self.CoreName = CoreName
        self.CollectType = CollectType
        self.RadarMode = RadarMode
        self.Classification = Classification
        self.CountryCodes = CountryCodes
        self.Parameters = Parameters
        super(CollectionInfoType, self).__init__(**kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class EXPLTBType(TREElement):
    def __init__(self, value):
        super(EXPLTBType, self).__init__()
        self.add_field('ANGLE_TO_NORTH', 's', 7, value)
        self.add_field('ANGLE_TO_NORTH_ACCY', 's', 6, value)
        self.add_field('SQUINT_ANGLE', 's', 7, value)
        self.add_field('SQUINT_ANGLE_ACCY', 's', 6, value)
        self.add_field('MODE', 's', 3, value)
        self.add_field('RESVD001', 's', 16, value)
        self.add_field('GRAZE_ANG', 's', 5, value)
        self.add_field('GRAZE_ANG_ACCY', 's', 5, value)
        self.add_field('SLOPE_ANG', 's', 5, value)
        self.add_field('POLAR', 's', 2, value)
        self.add_field('NSAMP', 's', 5, value)
        self.add_field('RESVD002', 's', 1, value)
        self.add_field('SEQ_NUM', 's', 1, value)
        self.add_field('PRIME_ID', 's', 12, value)
        self.add_field('PRIME_BE', 's', 15, value)
        self.add_field('RESVD003', 's', 1, value)
        self.add_field('N_SEC', 's', 2, value)
        self.add_field('IPR', 's', 2, value)


class EXPLTB(TREExtension):
    _tag_value = 'EXPLTB'
    _data_type = EXPLTBType



"""
The ImageFormationType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Optional, Dict, Tuple
from datetime import datetime, date

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, ParametersCollection
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    FloatDescriptor, IntegerDescriptor, IntegerListDescriptor, BooleanDescriptor, \
    ComplexDescriptor, DateTimeDescriptor, SerializableDescriptor, \
    SerializableListDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import DUAL_POLARIZATION_VALUES
from .RadarCollection import get_band_name
from .utils import polstring_version_required


class RcvChanProcType(Serializable):
    """The Received Processed Channels."""
    _fields = ('NumChanProc', 'PRFScaleFactor', 'ChanIndices')
    _required = ('NumChanProc', 'ChanIndices')
    _collections_tags = {
        'ChanIndices': {'array': False, 'child_tag': 'ChanIndex'}}
    _numeric_format = {'PRFScaleFactor': FLOAT_FORMAT}
    # descriptors
    NumChanProc = IntegerDescriptor(
        'NumChanProc', _required, strict=DEFAULT_STRICT,
        docstring='Number of receive data channels processed to form the image.')  # type: int
    PRFScaleFactor = FloatDescriptor(
        'PRFScaleFactor', _required, strict=DEFAULT_STRICT,
        docstring='Factor indicating the ratio of the effective PRF '
                  'to the actual PRF.')  # type: Optional[float]
    ChanIndices = IntegerListDescriptor(
        'ChanIndices', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Index of a data channel that was processed.')  # type: List[int]

    def __init__(
            self,
            NumChanProc: int = None,
            PRFScaleFactor: Optional[float] = None,
            ChanIndices: List[int] = None,
            **kwargs):
        """

        Parameters
        ----------
        NumChanProc : int
        PRFScaleFactor : float
        ChanIndices : List[int]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumChanProc = NumChanProc
        self.PRFScaleFactor = PRFScaleFactor
        self.ChanIndices = ChanIndices
        super(RcvChanProcType, self).__init__(**kwargs)


class TxFrequencyProcType(Serializable, Arrayable):
    """The transmit frequency range."""
    _fields = ('MinProc', 'MaxProc')
    _required = _fields
    _numeric_format = {'MinProc': '0.17E', 'MaxProc': '0.17E'}
    # descriptors
    MinProc = FloatDescriptor(
        'MinProc', _required, strict=DEFAULT_STRICT,
        docstring='The minimum transmit frequency processed to form the image, in Hz.')  # type: float
    MaxProc = FloatDescriptor(
        'MaxProc', _required, strict=DEFAULT_STRICT,
        docstring='The maximum transmit frequency processed to form the image, in Hz.')  # type: float

    def __init__(
            self,
            MinProc: float = None,
            MaxProc: float = None,
            **kwargs):
        """

        Parameters
        ----------
        MinProc : float
        MaxProc : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.MinProc, self.MaxProc = MinProc, MaxProc
        super(TxFrequencyProcType, self).__init__(**kwargs)

    @property
    def center_frequency(self) -> Optional[float]:
        """
        None|float: The center frequency.
        """

        if self.MinProc is None or self.MaxProc is None:
            return None
        return 0.5*(self.MinProc + self.MaxProc)

    @property
    def bandwidth(self) -> Optional[float]:
        """
        None|float: The bandwidth in Hz.
        """

        if self.MinProc is None or self.MaxProc is None:
            return None
        return self.MaxProc - self.MinProc

    def _apply_reference_frequency(
            self,
            reference_frequency: float):
        if self.MinProc is not None:
            self.MinProc += reference_frequency
        if self.MaxProc is not None:
            self.MaxProc += reference_frequency

    def _basic_validity_check(self) -> bool:
        condition = super(TxFrequencyProcType, self)._basic_validity_check()
        if self.MinProc is not None and self.MaxProc is not None and self.MaxProc < self.MinProc:
            self.log_validity_error(
                'Invalid frequency bounds MinProc ({}) > MaxProc ({})'.format(self.MinProc, self.MaxProc))
            condition = False
        return condition

    def get_band_name(self) -> str:
        """
        Gets the band name.

        Returns
        -------
        str
        """

        return get_band_name(self.center_frequency)

    def get_array(self, dtype=numpy.float64) -> numpy.ndarray:
        """
        Gets an array representation of the data.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array with appropriate entry order
        """

        return numpy.array([self.MinProc, self.MaxProc], dtype=dtype)

    @classmethod
    def from_array(cls, array: Union[numpy.ndarray, list, tuple]):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [MinProc, MaxProc]

        Returns
        -------
        LatLonType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(MinProc=array[0], MaxProc=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class ProcessingType(Serializable):
    """The transmit frequency range"""
    _fields = ('Type', 'Applied', 'Parameters')
    _required = ('Type', 'Applied')
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    Type = StringDescriptor(
        'Type', _required, strict=DEFAULT_STRICT,
        docstring='The processing type identifier.')  # type: str
    Applied = BooleanDescriptor(
        'Applied', _required, strict=DEFAULT_STRICT,
        docstring='Indicates whether the given processing type has been applied.')  # type: bool
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The parameters collection.')  # type: ParametersCollection

    def __init__(
            self,
            Type: str = None,
            Applied: bool = None,
            Parameters: Union[None, ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        Type : str
        Applied : bool
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Type = Type
        self.Applied = Applied
        self.Parameters = Parameters
        super(ProcessingType, self).__init__(**kwargs)


class DistortionType(Serializable):
    """Distortion"""
    _fields = (
        'CalibrationDate', 'A', 'F1', 'Q1', 'Q2', 'F2', 'Q3', 'Q4',
        'GainErrorA', 'GainErrorF1', 'GainErrorF2', 'PhaseErrorF1', 'PhaseErrorF2')
    _required = ('A', 'F1', 'Q1', 'Q2', 'F2', 'Q3', 'Q4')
    _numeric_format = {key: FLOAT_FORMAT for key in _fields[1:]}
    # descriptors
    CalibrationDate = DateTimeDescriptor(
        'CalibrationDate', _required, strict=DEFAULT_STRICT,
        docstring='The calibration date.')
    A = FloatDescriptor(
        'A', _required, strict=DEFAULT_STRICT,
        docstring='Absolute amplitude scale factor.')  # type: float
    # receive distortion matrix
    F1 = ComplexDescriptor(
        'F1', _required, strict=DEFAULT_STRICT,
        docstring='Receive distortion element (2,2).')  # type: complex
    Q1 = ComplexDescriptor(
        'Q1', _required, strict=DEFAULT_STRICT,
        docstring='Receive distortion element (1,2).')  # type: complex
    Q2 = ComplexDescriptor(
        'Q2', _required, strict=DEFAULT_STRICT,
        docstring='Receive distortion element (2,1).')  # type: complex
    # transmit distortion matrix
    F2 = ComplexDescriptor(
        'F2', _required, strict=DEFAULT_STRICT,
        docstring='Transmit distortion element (2,2).')  # type: complex
    Q3 = ComplexDescriptor(
        'Q3', _required, strict=DEFAULT_STRICT,
        docstring='Transmit distortion element (2,1).')  # type: complex
    Q4 = ComplexDescriptor(
        'Q4', _required, strict=DEFAULT_STRICT,
        docstring='Transmit distortion element (1,2).')  # type: complex
    # gain estimation error
    GainErrorA = FloatDescriptor(
        'GainErrorA', _required, strict=DEFAULT_STRICT,
        docstring='Gain estimation error standard deviation (in dB) for parameter A.')  # type: float
    GainErrorF1 = FloatDescriptor(
        'GainErrorF1', _required, strict=DEFAULT_STRICT,
        docstring='Gain estimation error standard deviation (in dB) for parameter F1.')  # type: float
    GainErrorF2 = FloatDescriptor(
        'GainErrorF2', _required, strict=DEFAULT_STRICT,
        docstring='Gain estimation error standard deviation (in dB) for parameter F2.')  # type: float
    PhaseErrorF1 = FloatDescriptor(
        'PhaseErrorF1', _required, strict=DEFAULT_STRICT,
        docstring='Phase estimation error standard deviation (in dB) for parameter F1.')  # type: float
    PhaseErrorF2 = FloatDescriptor(
        'PhaseErrorF2', _required, strict=DEFAULT_STRICT,
        docstring='Phase estimation error standard deviation (in dB) for parameter F2.')  # type: float

    def __init__(
            self,
            CalibrationDate: Union[None, numpy.datetime64, datetime, date, str] = None,
            A: float = None,
            F1: complex = None,
            Q1: complex = None,
            Q2: complex = None,
            F2: complex = None,
            Q3: complex = None,
            Q4: complex = None,
            GainErrorA: Optional[float] = None,
            GainErrorF1: Optional[float] = None,
            GainErrorF2: Optional[float] = None,
            PhaseErrorF1: Optional[float] = None,
            PhaseErrorF2: Optional[float] = None,
            **kwargs):
        """

        Parameters
        ----------
        CalibrationDate : numpy.datetime64|datetime|date|str
        A : float
        F1 : complex
        Q1 : complex
        Q2 : complex
        F2 : complex
        Q3 : complex
        Q4 : complex
        GainErrorA : float
        GainErrorF1 : float
        GainErrorF2 : float
        PhaseErrorF1 : float
        PhaseErrorF2 : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CalibrationDate = CalibrationDate
        self.A = A
        self.F1, self.Q1, self.Q2 = F1, Q1, Q2
        self.F2, self.Q3, self.Q4 = F2, Q3, Q4
        self.GainErrorA = GainErrorA
        self.GainErrorF1, self.GainErrorF2 = GainErrorF1, GainErrorF2
        self.PhaseErrorF1, self.PhaseErrorF2 = PhaseErrorF1, PhaseErrorF2
        super(DistortionType, self).__init__(**kwargs)


class PolarizationCalibrationType(Serializable):
    """The polarization calibration"""
    _fields = ('DistortCorrectApplied', 'Distortion')
    _required = _fields
    # descriptors
    DistortCorrectApplied = BooleanDescriptor(
        'DistortCorrectApplied', _required, strict=DEFAULT_STRICT,
        docstring='Indicates whether the polarization calibration has been applied.')  # type: bool
    Distortion = SerializableDescriptor(
        'Distortion', DistortionType, _required, strict=DEFAULT_STRICT,
        docstring='The distortion parameters.')  # type: DistortionType

    def __init__(
            self,
            DistortCorrectApplied: bool = None,
            Distortion: DistortionType = None,
            **kwargs):
        """

        Parameters
        ----------
        DistortCorrectApplied : bool
        Distortion : DistortionType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DistortCorrectApplied = DistortCorrectApplied
        self.Distortion = Distortion
        super(PolarizationCalibrationType, self).__init__(**kwargs)


class ImageFormationType(Serializable):
    """The image formation process parameters."""
    _fields = (
        'RcvChanProc', 'TxRcvPolarizationProc', 'TStartProc', 'TEndProc',
        'TxFrequencyProc', 'SegmentIdentifier', 'ImageFormAlgo', 'STBeamComp',
        'ImageBeamComp', 'AzAutofocus', 'RgAutofocus', 'Processings',
        'PolarizationCalibration')
    _required = (
        'RcvChanProc', 'TxRcvPolarizationProc', 'TStartProc', 'TEndProc', 'TxFrequencyProc',
        'ImageFormAlgo', 'STBeamComp', 'ImageBeamComp', 'AzAutofocus', 'RgAutofocus')
    _collections_tags = {'Processings': {'array': False, 'child_tag': 'Processing'}}
    _numeric_format = {'TStartProc': FLOAT_FORMAT, 'EndProc': FLOAT_FORMAT}
    # class variables
    _IMG_FORM_ALGO_VALUES = ('PFA', 'RMA', 'RGAZCOMP', 'OTHER')
    _ST_BEAM_COMP_VALUES = ('NO', 'GLOBAL', 'SV')
    _IMG_BEAM_COMP_VALUES = ('NO', 'SV')
    _AZ_AUTOFOCUS_VALUES = _ST_BEAM_COMP_VALUES
    _RG_AUTOFOCUS_VALUES = _ST_BEAM_COMP_VALUES
    # descriptors
    RcvChanProc = SerializableDescriptor(
        'RcvChanProc', RcvChanProcType, _required, strict=DEFAULT_STRICT,
        docstring='The received processed channels.')  # type: RcvChanProcType
    TxRcvPolarizationProc = StringEnumDescriptor(
        'TxRcvPolarizationProc', DUAL_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='The combined transmit/receive polarization processed to form the image.')  # type: str
    TStartProc = FloatDescriptor(
        'TStartProc', _required, strict=DEFAULT_STRICT,
        docstring='Earliest slow time value for data processed to form the image '
                  'from `CollectionStart`.')  # type: float
    TEndProc = FloatDescriptor(
        'TEndProc', _required, strict=DEFAULT_STRICT,
        docstring='Latest slow time value for data processed to form the image from `CollectionStart`.')  # type: float
    TxFrequencyProc = SerializableDescriptor(
        'TxFrequencyProc', TxFrequencyProcType, _required, strict=DEFAULT_STRICT,
        docstring='The range of transmit frequency processed to form the image.')  # type: TxFrequencyProcType
    SegmentIdentifier = StringDescriptor(
        'SegmentIdentifier', _required, strict=DEFAULT_STRICT,
        docstring='Identifier that describes the image that was processed. '
                  'Must be included when `SICD.RadarCollection.Area.Plane.SegmentList` is included.')  # type: str
    ImageFormAlgo = StringEnumDescriptor(
        'ImageFormAlgo', _IMG_FORM_ALGO_VALUES, _required, strict=DEFAULT_STRICT,
        docstring="""
        The image formation algorithm used:

        * `PFA` - Polar Format Algorithm

        * `RMA` - Range Migration (Omega-K, Chirp Scaling, Range-Doppler)

        * `RGAZCOMP` - Simple range, Doppler compression.

        """)  # type: str
    STBeamComp = StringEnumDescriptor(
        'STBeamComp', _ST_BEAM_COMP_VALUES, _required, strict=DEFAULT_STRICT,
        docstring="""
        Indicates if slow time beam shape compensation has been applied.

        * `NO` - No ST beam shape compensation.

        * `GLOBAL` - Global ST beam shape compensation applied.

        * `SV` - Spatially variant beam shape compensation applied.

        """)  # type: str
    ImageBeamComp = StringEnumDescriptor(
        'ImageBeamComp', _IMG_BEAM_COMP_VALUES, _required, strict=DEFAULT_STRICT,
        docstring="""
        Indicates if image domain beam shape compensation has been applied.

        * `NO` - No image domain beam shape compensation.

        * `SV` - Spatially variant image domain beam shape compensation applied.

        """)  # type: str
    AzAutofocus = StringEnumDescriptor(
        'AzAutofocus', _AZ_AUTOFOCUS_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Indicates if azimuth autofocus correction has been applied, with similar '
                  'interpretation as `STBeamComp`.')  # type: str
    RgAutofocus = StringEnumDescriptor(
        'RgAutofocus', _RG_AUTOFOCUS_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Indicates if range autofocus correction has been applied, with similar '
                  'interpretation as `STBeamComp`.')  # type: str
    Processings = SerializableListDescriptor(
        'Processings', ProcessingType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters to describe types of specific processing that may have been applied '
                  'such as additional compensations.')  # type: Optional[List[ProcessingType]]
    PolarizationCalibration = SerializableDescriptor(
        'PolarizationCalibration', PolarizationCalibrationType, _required, strict=DEFAULT_STRICT,
        docstring='The polarization calibration details.')  # type: Optional[PolarizationCalibrationType]

    def __init__(
            self,
            RcvChanProc: RcvChanProcType = None,
            TxRcvPolarizationProc: str = None,
            TStartProc: float = None,
            TEndProc: float = None,
            TxFrequencyProc: Union[TxFrequencyProcType, numpy.ndarray, list, tuple] = None,
            SegmentIdentifier: Optional[str] = None,
            ImageFormAlgo: str = None,
            STBeamComp: str = None,
            ImageBeamComp: str = None,
            AzAutofocus: str = None,
            RgAutofocus: str = None,
            Processings: Union[None, List[ProcessingType]] = None,
            PolarizationCalibration: Optional[PolarizationCalibrationType] = None,
            **kwargs):
        """

        Parameters
        ----------
        RcvChanProc : RcvChanProcType
        TxRcvPolarizationProc : str
        TStartProc : float
        TEndProc : float
        TxFrequencyProc : TxFrequencyProcType|numpy.ndarray|list|tuple
        SegmentIdentifier : None|str
        ImageFormAlgo : str
        STBeamComp : str
        ImageBeamComp :str
        AzAutofocus : str
        RgAutofocus : str
        Processings : None|List[ProcessingType]
        PolarizationCalibration : None|PolarizationCalibrationType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvChanProc = RcvChanProc
        self.TxRcvPolarizationProc = TxRcvPolarizationProc
        self.TStartProc, self.TEndProc = TStartProc, TEndProc
        if isinstance(TxFrequencyProc, (numpy.ndarray, list, tuple)) and len(TxFrequencyProc) >= 2:
            self.TxFrequencyProc = TxFrequencyProcType(MinProc=TxFrequencyProc[0], MaxProc=TxFrequencyProc[1])
        else:
            self.TxFrequencyProc = TxFrequencyProc
        self.SegmentIdentifier = SegmentIdentifier
        self.ImageFormAlgo = ImageFormAlgo
        self.STBeamComp, self.ImageBeamComp = STBeamComp, ImageBeamComp
        self.AzAutofocus, self.RgAutofocus = AzAutofocus, RgAutofocus
        self.Processings = Processings
        self.PolarizationCalibration = PolarizationCalibration
        super(ImageFormationType, self).__init__(**kwargs)

    def _basic_validity_check(self) -> bool:
        condition = super(ImageFormationType, self)._basic_validity_check()
        if self.TStartProc is not None and self.TEndProc is not None and self.TEndProc < self.TStartProc:
            self.log_validity_error(
                'Invalid time processing bounds TStartProc ({}) > TEndProc ({})'.format(
                    self.TStartProc, self.TEndProc))
            condition = False
        return condition

    def _derive_tx_frequency_proc(self, RadarCollection):
        """
        Populate a default for processed frequency values, based on the assumption that the entire
        transmitted bandwidth was processed. This is expected to be called by SICD parent.

        Parameters
        ----------
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType

        Returns
        -------
        None
        """

        if RadarCollection is not None and RadarCollection.TxFrequency is not None and \
                RadarCollection.TxFrequency.Min is not None and RadarCollection.TxFrequency.Max is not None:
            # this is based on the assumption that the entire transmitted bandwidth was processed.
            if self.TxFrequencyProc is None:
                self.TxFrequencyProc = TxFrequencyProcType(
                    MinProc=RadarCollection.TxFrequency.Min, MaxProc=RadarCollection.TxFrequency.Max)
                # how would it make sense to set only one end?
            elif self.TxFrequencyProc.MinProc is None:
                self.TxFrequencyProc.MinProc = RadarCollection.TxFrequency.Min
            elif self.TxFrequencyProc.MaxProc is None:
                self.TxFrequencyProc.MaxProc = RadarCollection.TxFrequency.Max

    def _apply_reference_frequency(self, reference_frequency: float):
        """
        If the reference frequency is used, adjust the necessary fields accordingly.
        Expected to be called by SICD parent.

        Parameters
        ----------
        reference_frequency : float
            The reference frequency.

        Returns
        -------
        None
        """

        if self.TxFrequencyProc is not None:
            # noinspection PyProtectedMember
            self.TxFrequencyProc._apply_reference_frequency(reference_frequency)

    def get_polarization(self) -> str:
        """
        Gets the transmit/receive polarization.

        Returns
        -------
        str
        """

        return self.TxRcvPolarizationProc if self.TxRcvPolarizationProc is not None else 'UNKNOWN'

    def get_polarization_abbreviation(self) -> str:
        """
        Gets the transmit/receive polarization abbreviation for the suggested name.

        Returns
        -------
        str
        """

        pol = self.TxRcvPolarizationProc
        if pol is None or pol in ('OTHER', 'UNKNOWN'):
            return 'UN'
        fp, sp = pol.split(':')
        return fp[0]+sp[0]

    def get_transmit_band_name(self) -> str:
        """
        Gets the transmit band name.

        Returns
        -------
        str
        """

        if self.TxFrequencyProc is not None:
            return self.TxFrequencyProc.get_band_name()
        else:
            return 'UN'

    def version_required(self) -> Tuple[int, int, int]:
        """
        What SICD version is required?

        Returns
        -------
        tuple
        """

        return polstring_version_required(self.TxRcvPolarizationProc)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MSTGTAType(TREElement):
    def __init__(self, value):
        super(MSTGTAType, self).__init__()
        self.add_field('TGTNUM', 's', 5, value)
        self.add_field('TGTID', 's', 12, value)
        self.add_field('TGTBE', 's', 15, value)
        self.add_field('TGTPRI', 's', 3, value)
        self.add_field('TGTREQ', 's', 12, value)
        self.add_field('TGTLTIOV', 's', 12, value)
        self.add_field('TGTTYPE', 's', 1, value)
        self.add_field('TGTCOLL', 's', 1, value)
        self.add_field('TGTCAT', 's', 5, value)
        self.add_field('TGTUTC', 's', 7, value)
        self.add_field('TGTELEV', 's', 6, value)
        self.add_field('TGTELEVUNIT', 's', 1, value)
        self.add_field('TGTLOC', 's', 21, value)


class MSTGTA(TREExtension):
    _tag_value = 'MSTGTA'
    _data_type = MSTGTAType



"""
The Compensated Phase History Data 1.0.1 definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Daniel Pressler, Valkyrie")

from xml.etree import ElementTree
from collections import OrderedDict
from typing import Union
import numpy

from sarpy.io.xml.base import Serializable, find_children, parse_xml_from_file, \
    parse_xml_from_string
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    StringDescriptor
from sarpy.io.complex.sicd_elements.MatchInfo import MatchInfoType

from .base import DEFAULT_STRICT
from .CollectionID import CollectionIDType
from .Global import GlobalType
from .SceneCoordinates import SceneCoordinatesType
from .Data import DataType
from .Channel import ChannelType
from .PVP import PVPType
from .SupportArray import SupportArrayType
from .Dwell import DwellType
from .ReferenceGeometry import ReferenceGeometryType
from .Antenna import AntennaType
from .TxRcv import TxRcvType
from .ErrorParameters import ErrorParametersType
from .ProductInfo import ProductInfoType
from .GeoInfo import GeoInfoType

from sarpy.io.phase_history.cphd_schema import get_urn_details, WRITABLE_VERSIONS, \
    get_namespace, get_default_tuple

#########
# Module variables
_CPHD_SPEC_DETAILS = {
    key: {'namespace': get_namespace(key), 'details': get_urn_details(key)}
    for key in WRITABLE_VERSIONS}

_CPHD_DEFAULT_TUPLE = get_default_tuple()
_CPHD_DEFAULT_VERSION = '{}.{}.{}'.format(*_CPHD_DEFAULT_TUPLE)
CPHD_SECTION_TERMINATOR = b'\f\n'


#########
# CPHD header object

def _parse_cphd_header_field(line):
    """
    Parse the CPHD header field, or return `None` as a termination signal.

    Parameters
    ----------
    line : bytes

    Returns
    -------
    None|(str, str)
    """

    if line.startswith(CPHD_SECTION_TERMINATOR):
        return None
    parts = line.split(b' := ')
    if len(parts) != 2:
        raise ValueError('Cannot extract CPHD header value from line {}'.format(line))
    fld = parts[0].strip().decode('utf-8')
    val = parts[1].strip().decode('utf-8')
    return fld, val


class CPHDHeaderBase(object):
    _fields = ()
    _required = ()

    def __init__(self, **kwargs):
        # intended as an abstract object
        pass

    @classmethod
    def from_file_object(cls, fi):
        """
        Extract the CPHD header object from a file opened in byte mode.
        This file object is assumed to be at the correct location for the
        CPHD header.

        Parameters
        ----------
        fi
            The open file object, which will be progressively read.

        Returns
        -------
        CPHDHeaderBase
        """

        the_dict = {}
        while True:
            line = fi.readline()
            res = _parse_cphd_header_field(line)
            if res is None:
                break
            else:
                fld, val = res
            if fld not in cls._fields:
                raise ValueError('Cannot extract CPHD header value from line {}'.format(line))
            the_dict[fld] = val
        return cls(**the_dict)


class CPHDHeader(CPHDHeaderBase):
    _fields = (
        'XML_BLOCK_SIZE', 'XML_BLOCK_BYTE_OFFSET', 'SUPPORT_BLOCK_SIZE', 'SUPPORT_BLOCK_BYTE_OFFSET',
        'PVP_BLOCK_SIZE', 'PVP_BLOCK_BYTE_OFFSET', 'SIGNAL_BLOCK_SIZE', 'SIGNAL_BLOCK_BYTE_OFFSET',
        'CLASSIFICATION', 'RELEASE_INFO')
    _required = (
        'XML_BLOCK_SIZE', 'XML_BLOCK_BYTE_OFFSET', 'PVP_BLOCK_SIZE', 'PVP_BLOCK_BYTE_OFFSET',
        'SIGNAL_BLOCK_SIZE', 'SIGNAL_BLOCK_BYTE_OFFSET', 'CLASSIFICATION', 'RELEASE_INFO')
    # descriptor
    XML_BLOCK_SIZE = IntegerDescriptor(
        'XML_BLOCK_SIZE', _required, strict=True,
        docstring='ize of the XML instance that describes the product in bytes. '
                  'Size does NOT include the 2 bytes of the section terminator.')  # type: int
    XML_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'XML_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the XML block in bytes.')  # type: int
    SUPPORT_BLOCK_SIZE = IntegerDescriptor(
        'SUPPORT_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the Support block in bytes. Note - If the Support block is omitted, this '
                  'is not included.')  # type: int
    SUPPORT_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'SUPPORT_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the Support block in bytes. Note - If the Support '
                  'block is omitted, this is not included.')  # type: int
    PVP_BLOCK_SIZE = IntegerDescriptor(
        'PVP_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the PVP block in bytes.')  # type: int
    PVP_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'PVP_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the PVP block in bytes.')  # type: int
    SIGNAL_BLOCK_SIZE = IntegerDescriptor(
        'SIGNAL_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the Signal block in bytes.')  # type: int
    SIGNAL_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'SIGNAL_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the Signal block in bytes.')  # type: int
    CLASSIFICATION = StringDescriptor(
        'CLASSIFICATION', _required, strict=True, default_value='UNCLASSIFIED',
        docstring='Product classification information that is human-readable.')  # type: str
    RELEASE_INFO = StringDescriptor(
        'RELEASE_INFO', _required, strict=True, default_value='UNRESTRICTED',
        docstring='Product release information that is human-readable.')  # type: str

    def __init__(self, XML_BLOCK_SIZE=None, XML_BLOCK_BYTE_OFFSET=None,
                 SUPPORT_BLOCK_SIZE=None, SUPPORT_BLOCK_BYTE_OFFSET=None,
                 PVP_BLOCK_SIZE=None, PVP_BLOCK_BYTE_OFFSET=None,
                 SIGNAL_BLOCK_SIZE=None, SIGNAL_BLOCK_BYTE_OFFSET=None,
                 CLASSIFICATION='UNCLASSIFIED', RELEASE_INFO='UNRESTRICTED',
                 use_version=None):
        self.XML_BLOCK_SIZE = XML_BLOCK_SIZE
        self.XML_BLOCK_BYTE_OFFSET = XML_BLOCK_BYTE_OFFSET
        self.SUPPORT_BLOCK_SIZE = SUPPORT_BLOCK_SIZE
        self.SUPPORT_BLOCK_BYTE_OFFSET = SUPPORT_BLOCK_BYTE_OFFSET
        self.PVP_BLOCK_SIZE = PVP_BLOCK_SIZE
        self.PVP_BLOCK_BYTE_OFFSET = PVP_BLOCK_BYTE_OFFSET
        self.SIGNAL_BLOCK_SIZE = SIGNAL_BLOCK_SIZE
        self.SIGNAL_BLOCK_BYTE_OFFSET = SIGNAL_BLOCK_BYTE_OFFSET
        self.CLASSIFICATION = CLASSIFICATION
        self.RELEASE_INFO = RELEASE_INFO
        self._use_version = _CPHD_DEFAULT_VERSION if use_version is None else use_version
        super(CPHDHeader, self).__init__()

    @property
    def use_version(self) -> str:
        return self._use_version

    def to_string(self):
        """
        Forms a CPHD file header string (not including the section terminator) from populated attributes.
        """
        return ('CPHD/{}\n'.format(self.use_version)
                + ''.join(["{} := {}\n".format(f, getattr(self, f))
                           for f in self._fields if getattr(self, f) is not None]))


class CPHDType(Serializable):
    """
    The Compensated Phase History Data definition.
    """

    _fields = (
        'CollectionID', 'Global', 'SceneCoordinates', 'Data', 'Channel', 'PVP',
        'SupportArray', 'Dwell', 'ReferenceGeometry', 'Antenna', 'TxRcv',
        'ErrorParameters', 'ProductInfo', 'GeoInfo', 'MatchInfo')
    _required = (
        'CollectionID', 'Global', 'SceneCoordinates', 'Data', 'Channel', 'PVP',
        'Dwell', 'ReferenceGeometry')
    _collections_tags = {'GeoInfo': {'array': 'False', 'child_tag': 'GeoInfo'}}
    # descriptors
    CollectionID = SerializableDescriptor(
        'CollectionID', CollectionIDType, _required, strict=DEFAULT_STRICT,
        docstring='General information about the collection.')  # type: CollectionIDType
    Global = SerializableDescriptor(
        'Global', GlobalType, _required, strict=DEFAULT_STRICT,
        docstring='Global parameters that apply to metadata components and CPHD '
                  'signal arrays.')  # type: GlobalType
    SceneCoordinates = SerializableDescriptor(
        'SceneCoordinates', SceneCoordinatesType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that define geographic coordinates for in the imaged '
                  'scene.')  # type: SceneCoordinatesType
    Data = SerializableDescriptor(
        'Data', DataType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe binary data components contained in '
                  'the product.')  # type: DataType
    Channel = SerializableDescriptor(
        'Channel', ChannelType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the data channels contained in the '
                  'product.')  # type: ChannelType
    PVP = SerializableDescriptor(
        'PVP', PVPType, _required, strict=DEFAULT_STRICT,
        docstring='Structure specifying the Per Vector parameters provided for '
                  'each channel of a given product.')  # type: PVPType
    SupportArray = SerializableDescriptor(
        'SupportArray', SupportArrayType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the binary support array(s) content and '
                  'grid coordinates.')  # type: Union[None, SupportArrayType]
    Dwell = SerializableDescriptor(
        'Dwell', DwellType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that specify the dwell time supported by the signal '
                  'arrays contained in the CPHD product.')  # type: DwellType
    ReferenceGeometry = SerializableDescriptor(
        'ReferenceGeometry', ReferenceGeometryType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the collection geometry for the reference '
                  'vector of the reference channel.')  # type: ReferenceGeometryType
    Antenna = SerializableDescriptor(
        'Antenna', AntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the transmit and receive antennas used '
                  'to collect the signal array(s).')  # type: Union[None, AntennaType]
    TxRcv = SerializableDescriptor(
        'TxRcv', TxRcvType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the transmitted waveform(s) and receiver configurations '
                  'used in the collection.')  # type: Union[None, TxRcvType]
    ErrorParameters = SerializableDescriptor(
        'ErrorParameters', ErrorParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the statistics of errors in measured or estimated parameters '
                  'that describe the collection.')  # type: Union[None, ErrorParametersType]
    ProductInfo = SerializableDescriptor(
        'ProductInfo', ProductInfoType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that provide general information about the CPHD product '
                  'and/or the derived products that may be created '
                  'from it.')  # type: Union[None, ProductInfoType]
    MatchInfo = SerializableDescriptor(
        'MatchInfo', MatchInfoType, _required, strict=DEFAULT_STRICT,
        docstring='Information about other collections that are matched to the collection from which '
                  'this CPHD product was generated.')  # type: Union[None, MatchInfoType]

    def __init__(self, CollectionID=None, Global=None, SceneCoordinates=None, Data=None,
                 Channel=None, PVP=None, SupportArray=None, Dwell=None, ReferenceGeometry=None,
                 Antenna=None, TxRcv=None, ErrorParameters=None, ProductInfo=None,
                 GeoInfo=None, MatchInfo=None, **kwargs):
        """

        Parameters
        ----------
        CollectionID : CollectionIDType
        Global : GlobalType
        SceneCoordinates : SceneCoordinatesType
        Data : DataType
        Channel : ChannelType
        PVP : PVPType
        SupportArray : None|SupportArrayType
        Dwell : DwellType
        ReferenceGeometry : ReferenceGeometryType
        Antenna : None|AntennaType
        TxRcv : None|TxRcvType
        ErrorParameters : None|ErrorParametersType
        ProductInfo : None|ProductInfoType
        GeoInfo : None|List[GeoInfoType]|GeoInfoType
        MatchInfo : None|MatchInfoType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectionID = CollectionID
        self.Global = Global
        self.SceneCoordinates = SceneCoordinates
        self.Data = Data
        self.Channel = Channel
        self.PVP = PVP
        self.SupportArray = SupportArray
        self.Dwell = Dwell
        self.ReferenceGeometry = ReferenceGeometry
        self.Antenna = Antenna
        self.TxRcv = TxRcv
        self.ErrorParameters = ErrorParameters
        self.ProductInfo = ProductInfo
        self.MatchInfo = MatchInfo

        self._GeoInfo = []
        if GeoInfo is None:
            pass
        elif isinstance(GeoInfo, GeoInfoType):
            self.addGeoInfo(GeoInfo)
        elif isinstance(GeoInfo, (list, tuple)):
            for el in GeoInfo:
                self.addGeoInfo(el)
        else:
            raise ValueError('GeoInfo got unexpected type {}'.format(type(GeoInfo)))

        super(CPHDType, self).__init__(**kwargs)

    @property
    def GeoInfo(self):
        """
        List[GeoInfoType]: Parameters that describe a geographic feature.
        """

        return self._GeoInfo

    def getGeoInfo(self, key):
        """
        Get GeoInfo(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfo if entry.name == key]

    def addGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfo list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfo', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfo.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfo', ns_key)
        kwargs['GeoInfo'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(CPHDType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(CPHDType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity,
            strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self._GeoInfo is not None and len(self._GeoInfo) > 0:
            for entry in self._GeoInfo:
                entry.to_node(doc, 'GeoInfo', ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(CPHDType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if len(self.GeoInfo) > 0:
            out['GeoInfo'] = [entry.to_dict(
                check_validity=check_validity, strict=strict) for entry in self._GeoInfo]
        return out

    def to_xml_bytes(self, urn=None, tag='CPHD', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = get_namespace(_CPHD_DEFAULT_VERSION)
        return super(CPHDType, self).to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='CPHD', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def version_required(self):
        """
        What CPHD version is required for valid support?

        Returns
        -------
        tuple
        """

        required = (1, 0, 1)
        for fld in self._fields:
            val = getattr(self, fld)
            if val is not None and hasattr(val, 'version_required'):
                required = max(required, val.version_required())
        return required

    def make_file_header(self, xml_offset=1024, use_version=None):
        """
        Forms a CPHD file header consistent with the information in the Data and CollectionID nodes.

        Parameters
        ----------
        xml_offset : int, optional
            Offset in bytes to the first byte of the XML block. If the provided value is not large enough to account for
            the length of the file header string, a larger value is chosen.
        use_version : None|str
            What version to use?

        Returns
        -------
        header : CPHDHeader
        """

        kwargs = OrderedDict()
        kwargs['use_version'] = _CPHD_DEFAULT_VERSION if use_version is None else use_version

        def _align(val):
            align_to = 64
            return int(numpy.ceil(float(val)/align_to)*align_to)

        kwargs['XML_BLOCK_SIZE'] = len(self.to_xml_string())
        kwargs['XML_BLOCK_BYTE_OFFSET'] = xml_offset
        block_end = kwargs['XML_BLOCK_BYTE_OFFSET'] + kwargs['XML_BLOCK_SIZE'] + len(CPHD_SECTION_TERMINATOR)

        if self.Data.NumSupportArrays > 0:
            kwargs['SUPPORT_BLOCK_SIZE'] = self.Data.calculate_support_block_size()
            kwargs['SUPPORT_BLOCK_BYTE_OFFSET'] = _align(block_end)
            block_end = kwargs['SUPPORT_BLOCK_BYTE_OFFSET'] + kwargs['SUPPORT_BLOCK_SIZE']

        kwargs['PVP_BLOCK_SIZE'] = self.Data.calculate_pvp_block_size()
        kwargs['PVP_BLOCK_BYTE_OFFSET'] = _align(block_end)
        block_end = kwargs['PVP_BLOCK_BYTE_OFFSET'] + kwargs['PVP_BLOCK_SIZE']

        kwargs['SIGNAL_BLOCK_SIZE'] = self.Data.calculate_signal_block_size()
        kwargs['SIGNAL_BLOCK_BYTE_OFFSET'] = _align(block_end)
        kwargs['CLASSIFICATION'] = self.CollectionID.Classification
        kwargs['RELEASE_INFO'] = self.CollectionID.ReleaseInfo

        header = CPHDHeader(**kwargs)
        header_str = header.to_string()
        min_xml_offset = len(header_str) + len(CPHD_SECTION_TERMINATOR)
        if kwargs['XML_BLOCK_BYTE_OFFSET'] < min_xml_offset:
            header = self.make_file_header(xml_offset=_align(min_xml_offset + 32), use_version=use_version)

        return header

    def get_pvp_dtype(self):
        """
        Gets the dtype for the corresponding PVP structured array. Note that they
        must all have homogeneous dtype.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        if self.PVP is None:
            raise ValueError('No PVP defined.')
        return self.PVP.get_vector_dtype()

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the cphd object from a stand-alone xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        CPHDType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the cphd object from an xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        CPHDType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class ST(TREElement):
    def __init__(self, value):
        super(ST, self).__init__()
        self.add_field('SECTITLE', 's', 40, value)
        self.add_field('PPNUM', 's', 5, value)
        self.add_field('TPP', 's', 3, value)


class RO(TREElement):
    def __init__(self, value):
        super(RO, self).__init__()
        self.add_field('REQORG', 's', 64, value)


class KW(TREElement):
    def __init__(self, value):
        super(KW, self).__init__()
        self.add_field('KEYWORD', 's', 255, value)


class AR(TREElement):
    def __init__(self, value):
        super(AR, self).__init__()
        self.add_field('ASSRPT', 's', 20, value)


class AT(TREElement):
    def __init__(self, value):
        super(AT, self).__init__()
        self.add_field('ATEXT', 's', 255, value)


class PIAPRCType(TREElement):
    def __init__(self, value):
        super(PIAPRCType, self).__init__()
        self.add_field('ACCID', 's', 64, value)
        self.add_field('FMCTL', 's', 32, value)
        self.add_field('SDET', 's', 1, value)
        self.add_field('PCODE', 's', 2, value)
        self.add_field('PSUBE', 's', 6, value)
        self.add_field('PIDNM', 's', 20, value)
        self.add_field('PNAME', 's', 10, value)
        self.add_field('MAKER', 's', 2, value)
        self.add_field('CTIME', 's', 14, value)
        self.add_field('MAPID', 's', 40, value)
        self.add_field('STREP', 'd', 2, value)
        self.add_loop('STs', self.STREP, ST, value)
        self.add_field('ROREP', 'd', 2, value)
        self.add_loop('ROs', self.ROREP, RO, value)
        self.add_field('KWREP', 'd', 2, value)
        self.add_loop('KWs', self.KWREP, KW, value)
        self.add_field('ARREP', 'd', 2, value)
        self.add_loop('ARs', self.ARREP, AR, value)
        self.add_field('ATREP', 'd', 2, value)
        self.add_loop('ATs', self.ATREP, AT, value)


class PIAPRC(TREExtension):
    _tag_value = 'PIAPRC'
    _data_type = PIAPRCType



"""
Functionality for reading ICEYE complex data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import os
from typing import Union, Tuple, Sequence, Optional

import numpy
from numpy.polynomial import polynomial
from scipy.constants import speed_of_light

from sarpy.io.complex.nisar import _stringify
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import Poly2DType, Poly1DType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, \
    RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    ChanParametersType, WaveformParametersType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Position import PositionType, XYZPolyType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType
from sarpy.io.complex.utils import fit_position_xvalidation, two_dim_poly_fit

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.data_segment import HDF5DatasetSegment, BandAggregateSegment
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.utils import get_seconds, parse_timestring, is_file_like, is_hdf5, h5py

logger = logging.getLogger(__name__)


def _parse_time(input_str: Union[bytes, str]) -> numpy.datetime64:
    """
    Parse the timestring.

    Parameters
    ----------
    input_str : bytes|str

    Returns
    -------
    numpy.datetime64
    """

    return parse_timestring(_stringify(input_str), precision='us')


class ICEYEDetails(object):
    """
    Parses and converts the ICEYE metadata.
    """
    __slots__ = ('_file_name', )

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        if h5py is None:
            raise ImportError("Can't read ICEYE files, because the h5py dependency is missing.")

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))

        with h5py.File(file_name, 'r') as hf:
            if 's_q' not in hf or 's_i' not in hf:
                raise SarpyIOError(
                    'The hdf file does not have the real (s_q) or imaginary dataset (s_i).')
            if 'satellite_name' not in hf:
                raise SarpyIOError('The hdf file does not have the satellite_name dataset.')
            if 'product_name' not in hf:
                raise SarpyIOError('The hdf file does not have the product_name dataset.')

        self._file_name = file_name

    @property
    def file_name(self) -> str:
        """
        str: the file name
        """

        return self._file_name

    def get_sicd(self) -> (SICDType, Optional[Tuple[int, ...]], Tuple[int, ...]):
        """
        Gets the SICD structure and associated details for constructing the data segment.

        Returns
        -------
        sicd : SICDType
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : Tuple[int, ...]
        """

        def get_collection_info() -> CollectionInfoType:

            mode_id_type_map = {
                "SpotlightExtendedArea": "DYNAMIC STRIPMAP", }
            mode_id = _stringify(hf['product_type'][()])
            mode_type = mode_id_type_map.get(mode_id, _stringify(hf['acquisition_mode'][()]).upper())

            return CollectionInfoType(
                CollectorName=_stringify(hf['satellite_name'][()]),
                CoreName=_stringify(hf['product_name'][()]),
                CollectType='MONOSTATIC',
                Classification='UNCLASSIFIED',
                RadarMode=RadarModeType(
                    ModeType=mode_type,
                    ModeID=mode_id))

        def get_image_creation() -> ImageCreationType:
            from sarpy.__about__ import __version__
            return ImageCreationType(
                Application='ICEYE_P_{}'.format(hf['processor_version'][()]),
                DateTime=_parse_time(hf['processing_time'][()]),
                Site='Unknown',
                Profile='sarpy {}'.format(__version__))

        def get_image_data() -> ImageDataType:

            samp_prec = _stringify(hf['sample_precision'][()])
            if samp_prec.upper() == 'INT16':
                pixel_type = 'RE16I_IM16I'
            elif samp_prec.upper() == 'FLOAT32':
                pixel_type = 'RE32F_IM32F'
            else:
                raise ValueError('Got unhandled sample precision {}'.format(samp_prec))

            num_rows = int(number_of_range_samples)
            num_cols = int(number_of_azimuth_samples)
            scp_row = int(coord_center[0]) - 1
            scp_col = int(coord_center[1]) - 1
            if 0 < scp_col < num_rows-1:
                if look_side == 'left':
                    scp_col = num_cols - scp_col - 1
            else:
                # early ICEYE processing bug led to nonsensical SCP
                scp_col = int(num_cols/2.0)

            return ImageDataType(
                PixelType=pixel_type,
                NumRows=num_rows,
                NumCols=num_cols,
                FirstRow=0,
                FirstCol=0,
                FullImage=(num_rows, num_cols),
                SCPPixel=(scp_row, scp_col))

        def get_geo_data() -> GeoDataType:
            # NB: the remainder will be derived.
            return GeoDataType(
                SCP=SCPType(
                    LLH=[coord_center[2], coord_center[3], avg_scene_height]))

        def get_timeline() -> TimelineType:
            acq_prf = hf['acquisition_prf'][()]
            return TimelineType(
                CollectStart=start_time,
                CollectDuration=duration,
                IPP=[IPPSetType(index=0, TStart=0, TEnd=duration,
                                IPPStart=0, IPPEnd=int(round(acq_prf*duration) - 1),
                                IPPPoly=[0, acq_prf]), ])

        def get_position() -> PositionType:
            times_str = hf['state_vector_time_utc'][:, 0]
            times = numpy.zeros((times_str.shape[0], ), dtype='float64')
            positions = numpy.zeros((times.size, 3), dtype='float64')
            velocities = numpy.zeros((times.size, 3), dtype='float64')
            for i, entry in enumerate(times_str):
                times[i] = get_seconds(_parse_time(entry), start_time, precision='us')
            positions[:, 0], positions[:, 1], positions[:, 2] = hf['posX'][:], hf['posY'][:], hf['posZ'][:]
            velocities[:, 0], velocities[:, 1], velocities[:, 2] = hf['velX'][:], hf['velY'][:], hf['velZ'][:]
            # fir the the position polynomial using cross validation
            P_x, P_y, P_z = fit_position_xvalidation(times, positions, velocities, max_degree=8)
            return PositionType(ARPPoly=XYZPolyType(X=P_x, Y=P_y, Z=P_z))

        def get_radar_collection() -> RadarCollectionType:
            return RadarCollectionType(
                TxPolarization=tx_pol,
                TxFrequency=(min_freq, max_freq),
                Waveform=[WaveformParametersType(TxFreqStart=min_freq,
                                                 TxRFBandwidth=tx_bandwidth,
                                                 TxPulseLength=hf['chirp_duration'][()],
                                                 ADCSampleRate=hf['range_sampling_rate'][()],
                                                 RcvDemodType='CHIRP',
                                                 RcvFMRate=0,
                                                 index=1)],
                RcvChannels=[ChanParametersType(TxRcvPolarization=polarization,
                                                index=1)])

        def get_image_formation() -> ImageFormationType:
            return ImageFormationType(
                TxRcvPolarizationProc=polarization,
                ImageFormAlgo='RMA',
                TStartProc=0,
                TEndProc=duration,
                TxFrequencyProc=(min_freq, max_freq),
                STBeamComp='NO',
                ImageBeamComp='SV',
                AzAutofocus='NO',
                RgAutofocus='NO',
                RcvChanProc=RcvChanProcType(NumChanProc=1, PRFScaleFactor=1, ChanIndices=[1, ]),)

        def get_radiometric() -> RadiometricType:
            return RadiometricType(BetaZeroSFPoly=[[float(hf['calibration_factor'][()]), ], ])

        def calculate_drate_sf_poly() -> (numpy.ndarray, numpy.ndarray):
            r_ca_coeffs = numpy.array([r_ca_scp, 1], dtype='float64')
            dop_rate_coeffs = hf['doppler_rate_coeffs'][:]
            # Prior to ICEYE 1.14 processor, absolute value of Doppler rate was
            # provided, not true Doppler rate. Doppler rate should always be negative
            if dop_rate_coeffs[0] > 0:
                dop_rate_coeffs *= -1
            dop_rate_poly = Poly1DType(Coefs=dop_rate_coeffs)
            # now adjust to create
            t_drate_ca_poly = dop_rate_poly.shift(
                t_0=zd_ref_time - rg_time_scp,
                alpha=2/speed_of_light, return_poly=False)
            return t_drate_ca_poly, \
                -polynomial.polymul(t_drate_ca_poly, r_ca_coeffs)*speed_of_light/(2*center_freq*vm_ca_sq)

        def calculate_doppler_polys() -> (numpy.ndarray, numpy.ndarray):
            # define and fit the time coa array
            if collect_info.RadarMode.ModeType == 'SPOTLIGHT':
                coa_time = duration / 2
                look = 1 if look_side == 'left' else -1
                alpha = 2.0/speed_of_light
                pos = position.ARPPoly(coa_time)
                vel = position.ARPPoly.derivative_eval(coa_time)
                speed = numpy.linalg.norm(vel)
                vel_hat = vel / speed
                los = geo_data.SCP.ECF.get_array() - pos

                t_time_coa_coeffs = [[coa_time, ], ]
                t_dop_centroid_coeffs = numpy.zeros((2, 2), dtype=numpy.float64)
                t_dop_centroid_coeffs[0, 1] = -look*center_freq*alpha*speed/r_ca_scp
                t_dop_centroid_coeffs[1, 1] = look*center_freq*alpha*speed/(r_ca_scp**2)
                t_dop_centroid_coeffs[:, 0] = -look*(t_dop_centroid_coeffs[:, 1]*numpy.dot(los, vel_hat))
            else:
                # extract doppler centroid coefficients
                dc_estimate_coeffs = hf['dc_estimate_coeffs'][:]
                dc_time_str = hf['dc_estimate_time_utc'][:, 0]
                dc_zd_times = numpy.zeros((dc_time_str.shape[0], ), dtype='float64')
                for i, entry in enumerate(dc_time_str):
                    dc_zd_times[i] = get_seconds(_parse_time(entry), start_time, precision='us')
                # create a sampled doppler centroid
                samples = 49  # copied from corresponding matlab, we just need enough for appropriate refitting
                # create doppler time samples
                diff_time_rg = first_pixel_time - zd_ref_time + \
                    numpy.linspace(0, number_of_range_samples/range_sampling_rate, samples)
                # doppler centroid samples definition
                dc_sample_array = numpy.zeros((samples, dc_zd_times.size), dtype='float64')
                for i, coeffs in enumerate(dc_estimate_coeffs):
                    dc_sample_array[:, i] = polynomial.polyval(diff_time_rg, coeffs)
                # create arrays for range/azimuth from scp in meters
                azimuth_scp_m, range_scp_m = numpy.meshgrid(
                    col_ss*(dc_zd_times - zd_time_scp)/ss_zd_s,
                    (diff_time_rg + zd_ref_time - rg_time_scp)*speed_of_light/2)

                # fit the doppler centroid sample array
                x_order = min(3, range_scp_m.shape[0]-1)
                y_order = min(3, range_scp_m.shape[1]-1)

                t_dop_centroid_coeffs, residuals, rank, sing_values = two_dim_poly_fit(
                    range_scp_m, azimuth_scp_m, dc_sample_array, x_order=x_order, y_order=y_order,
                    x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
                logger.info(
                    'The dop_centroid_poly fit details:\n\troot mean square '
                    'residuals = {}\n\trank = {}\n\tsingular values = {}'.format(
                        residuals, rank, sing_values))

                doppler_rate_sampled = polynomial.polyval(azimuth_scp_m, drate_ca_poly)
                time_coa = dc_zd_times + dc_sample_array/doppler_rate_sampled
                t_time_coa_coeffs, residuals, rank, sing_values = two_dim_poly_fit(
                    range_scp_m, azimuth_scp_m, time_coa, x_order=x_order, y_order=y_order,
                    x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
                logger.info(
                    'The time_coa_poly fit details:\n\troot mean square '
                    'residuals = {}\n\trank = {}\n\tsingular values = {}'.format(
                        residuals, rank, sing_values))

            return t_dop_centroid_coeffs, t_time_coa_coeffs

        def get_rma() -> RMAType:
            if collect_info.RadarMode.ModeType == 'SPOTLIGHT':
                dop_centroid_poly = None
                dop_centroid_coa = None
            else:
                dop_centroid_poly = Poly2DType(Coefs=dop_centroid_poly_coeffs)
                dop_centroid_coa = True
            # NB: DRateSFPoly is defined as a function of only range - reshape appropriately
            inca = INCAType(
                R_CA_SCP=r_ca_scp,
                FreqZero=center_freq,
                DRateSFPoly=Poly2DType(Coefs=numpy.reshape(drate_sf_poly_coefs, (-1, 1))),
                DopCentroidPoly=dop_centroid_poly,
                DopCentroidCOA=dop_centroid_coa,
                TimeCAPoly=Poly1DType(Coefs=time_ca_poly_coeffs))
            return RMAType(
                RMAlgoType='OMEGA_K',
                INCA=inca)

        def get_grid() -> GridType:
            time_coa_poly = Poly2DType(Coefs=time_coa_poly_coeffs)

            row_win = _stringify(hf['window_function_range'][()])
            if row_win == 'NONE':
                row_win = 'UNIFORM'
            row = DirParamType(
                SS=row_ss,
                Sgn=-1,
                KCtr=2*center_freq/speed_of_light,
                ImpRespBW=2*tx_bandwidth/speed_of_light,
                DeltaKCOAPoly=Poly2DType(Coefs=[[0, ], ]),
                WgtType=WgtTypeType(WindowName=row_win))
            col_win = _stringify(hf['window_function_azimuth'][()])
            if col_win == 'NONE':
                col_win = 'UNIFORM'
            col = DirParamType(
                SS=col_ss,
                Sgn=-1,
                KCtr=0,
                ImpRespBW=col_imp_res_bw,
                WgtType=WgtTypeType(WindowName=col_win),
                DeltaKCOAPoly=Poly2DType(Coefs=dop_centroid_poly_coeffs*ss_zd_s/col_ss))
            return GridType(
                Type='RGZERO',
                ImagePlane='SLANT',
                TimeCOAPoly=time_coa_poly,
                Row=row,
                Col=col)

        def correct_scp() -> None:
            scp_pixel = sicd.ImageData.SCPPixel.get_array()
            scp_ecf = sicd.project_image_to_ground(scp_pixel, projection_type='HAE')
            sicd.update_scp(scp_ecf, coord_system='ECF')

        with h5py.File(self._file_name, 'r') as hf:
            # some common use variables
            look_side = _stringify(hf['look_side'][()])
            coord_center = hf['coord_center'][:]
            avg_scene_height = float(hf['avg_scene_height'][()])
            start_time = _parse_time(hf['acquisition_start_utc'][()])
            end_time = _parse_time(hf['acquisition_end_utc'][()])
            duration = get_seconds(end_time, start_time, precision='us')

            center_freq = float(hf['carrier_frequency'][()])
            tx_bandwidth = float(hf['chirp_bandwidth'][()])
            min_freq = center_freq-0.5*tx_bandwidth
            max_freq = center_freq+0.5*tx_bandwidth

            pol_temp = _stringify(hf['polarization'][()])
            tx_pol = pol_temp[0]
            rcv_pol = pol_temp[1]
            polarization = tx_pol + ':' + rcv_pol

            first_pixel_time = float(hf['first_pixel_time'][()])
            near_range = first_pixel_time*speed_of_light/2
            number_of_range_samples = float(hf['number_of_range_samples'][()])
            number_of_azimuth_samples = float(hf['number_of_azimuth_samples'][()])
            range_sampling_rate = float(hf['range_sampling_rate'][()])
            row_ss = speed_of_light/(2*range_sampling_rate)

            # define the sicd elements
            collect_info = get_collection_info()
            image_creation = get_image_creation()
            image_data = get_image_data()
            geo_data = get_geo_data()
            timeline = get_timeline()
            position = get_position()
            radar_collection = get_radar_collection()
            image_formation = get_image_formation()
            radiometric = get_radiometric()

            # calculate some zero doppler parameters
            ss_zd_s = float(hf['azimuth_time_interval'][()])
            if look_side == 'left':
                ss_zd_s *= -1
                zero_doppler_left = _parse_time(hf['zerodoppler_end_utc'][()])
            else:
                zero_doppler_left = _parse_time(hf['zerodoppler_start_utc'][()])
            dop_bw = hf['total_processed_bandwidth_azimuth'][()]
            zd_time_scp = get_seconds(zero_doppler_left, start_time, precision='us') + \
                image_data.SCPPixel.Col*ss_zd_s
            zd_ref_time = first_pixel_time + number_of_range_samples/(2*range_sampling_rate)
            vel_scp = position.ARPPoly.derivative_eval(zd_time_scp, der_order=1)
            vm_ca_sq = numpy.sum(vel_scp*vel_scp)
            rg_time_scp = first_pixel_time + image_data.SCPPixel.Row/range_sampling_rate
            r_ca_scp = rg_time_scp*speed_of_light/2
            # calculate the doppler rate sf polynomial
            drate_ca_poly, drate_sf_poly_coefs = calculate_drate_sf_poly()

            # calculate some doppler dependent grid parameters
            col_ss = float(numpy.sqrt(vm_ca_sq)*abs(ss_zd_s)*drate_sf_poly_coefs[0])
            col_imp_res_bw = dop_bw*abs(ss_zd_s)/col_ss
            time_ca_poly_coeffs = [zd_time_scp, ss_zd_s/col_ss]

            # calculate the doppler polynomials
            dop_centroid_poly_coeffs, time_coa_poly_coeffs = calculate_doppler_polys()

            # finish definition of sicd elements
            rma = get_rma()
            grid = get_grid()
            sicd = SICDType(
                CollectionInfo=collect_info,
                ImageCreation=image_creation,
                ImageData=image_data,
                GeoData=geo_data,
                Timeline=timeline,
                Position=position,
                RadarCollection=radar_collection,
                ImageFormation=image_formation,
                Radiometric=radiometric,
                RMA=rma,
                Grid=grid)
        # adjust the scp location
        correct_scp()
        # derive sicd fields
        sicd.derive()
        # TODO: RNIIRS?

        reverse_axes = (0, ) if look_side == 'left' else None
        transpose_axes = (1, 0)
        # noinspection PyTypeChecker
        return sicd, reverse_axes, transpose_axes


def get_iceye_data_segment(
        file_name: str,
        reverse_axes: Union[None, int, Sequence[int]],
        transpose_axes: Union[None, Tuple[int, ...]],
        real_group: str = 's_i',
        imaginary_grop: str = 's_q') -> BandAggregateSegment:
    real_dataset = HDF5DatasetSegment(
        file_name, real_group, reverse_axes=reverse_axes, transpose_axes=transpose_axes, close_file=True)
    imag_dataset = HDF5DatasetSegment(
        file_name, imaginary_grop, reverse_axes=reverse_axes, transpose_axes=transpose_axes, close_file=True)
    return BandAggregateSegment(
        (real_dataset, imag_dataset), band_dimension=2,
        formatted_dtype='complex64', formatted_shape=real_dataset.formatted_shape,
        format_function=ComplexFormatFunction(real_dataset.formatted_dtype, order='IQ', band_dimension=-1))


class ICEYEReader(SICDTypeReader):
    """
    An ICEYE SLC reader implementation.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_iceye_details', )

    def __init__(self, iceye_details):
        """

        Parameters
        ----------
        iceye_details : str|ICEYEDetails
            file name or ICEYEDetails object
        """

        if isinstance(iceye_details, str):
            iceye_details = ICEYEDetails(iceye_details)
        if not isinstance(iceye_details, ICEYEDetails):
            raise TypeError('The input argument for a ICEYEReader must be a '
                            'filename or ICEYEDetails object')
        self._iceye_details = iceye_details
        sicd, reverse_axes, transpose_axes = iceye_details.get_sicd()
        data_segment = get_iceye_data_segment(iceye_details.file_name, reverse_axes, transpose_axes)

        SICDTypeReader.__init__(self, data_segment, sicd, close_segments=True)
        self._check_sizes()

    @property
    def iceye_details(self) -> ICEYEDetails:
        """
        ICEYEDetails: The ICEYE details object.
        """

        return self._iceye_details

    @property
    def file_name(self):
        return self.iceye_details.file_name


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Union[None, ICEYEReader]:
    """
    Tests whether a given file_name corresponds to a ICEYE file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str|BinaryIO
        the file_name to check

    Returns
    -------
    None|ICEYEReader
        `ICEYEReader` instance if ICEYE file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    if not is_hdf5(file_name):
        return None

    if h5py is None:
        return None

    try:
        iceye_details = ICEYEDetails(file_name)
        logger.info('File {} is determined to be a ICEYE file.'.format(file_name))
        return ICEYEReader(iceye_details)
    except SarpyIOError:
        return None



"""
Radar Generalized Image Quality Equation (RGIQE) calculation(s) and tools for
application to SICD structures and files.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union, Tuple, Dict, Optional

import numpy

from sarpy.io.complex.base import SICDTypeReader, FlatSICDReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.processing.sicd.windows import get_hamming_broadening_factor
from sarpy.processing.sicd.normalize_sicd import sicd_degrade_reweight, is_uniform_weight
from sarpy.io.complex.converter import open_complex

logger = logging.getLogger(__name__)

RNIIRS_FIT_PARAMETERS = numpy.array([3.4761, 0.4357], dtype='float64')
"""
The RNIIRS calculation parameters determined by empirical data fit, 
parameters updated on 2022-02-01
"""


#####################
# methods for extracting necessary information from the sicd structure

def _verify_sicd_with_noise(sicd: SICDType) -> None:
    """
    Verify that the sicd is appropriately populated with noise.

    Parameters
    ----------
    sicd : SICDType
    """

    if sicd.Radiometric is None:
        raise ValueError(
            'Radiometric is not populated,\n\t'
            'so no noise estimate can be derived.')
    if sicd.Radiometric.SigmaZeroSFPoly is None:
        raise ValueError(
            'Radiometric.SigmaZeroSFPoly is not populated,\n\t'
            'so no sigma0 noise estimate can be derived.')
    if sicd.Radiometric.NoiseLevel is None:
        raise ValueError(
            'Radiometric.NoiseLevel is not populated,\n\t'
            'so no noise estimate can be derived.')
    if sicd.Radiometric.NoiseLevel.NoiseLevelType != 'ABSOLUTE':
        raise ValueError(
            'Radiometric.NoiseLevel.NoiseLevelType is not `ABSOLUTE``,\n\t'
            'so no noise estimate can be derived.')


def get_sigma0_noise(sicd: SICDType) -> float:
    """
    Calculate the absolute noise estimate, in sigma0 power units.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    float
    """

    _verify_sicd_with_noise(sicd)
    noise = sicd.Radiometric.NoiseLevel.NoisePoly[0, 0]  # this is in db
    noise = numpy.exp(numpy.log(10)*0.1*noise)  # this is absolute

    # convert to SigmaZero value
    noise *= sicd.Radiometric.SigmaZeroSFPoly[0, 0]
    return noise


def get_default_signal_estimate(sicd: SICDType) -> float:
    """
    Gets default signal for use in the RNIIRS calculation. This will be
    1.0 for copolar (or unknown) collections, and 0.25 for cross-pole
    collections.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    float
    """

    if sicd.ImageFormation is None or sicd.ImageFormation.TxRcvPolarizationProc is None:
        return 1.0

    # use 1.0 for co-polar collection and 0.25 from cross-polar collection
    pol = sicd.ImageFormation.TxRcvPolarizationProc
    if pol is None or ':' not in pol:
        return 1.0

    pols = pol.split(':')

    return 1.0 if pols[0] == pols[1] else 0.25


def get_bandwidth_area(sicd: SICDType) -> float:
    """
    Calculate the bandwidth area.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    float
    """

    return abs(
        sicd.Grid.Row.ImpRespBW *
        sicd.Grid.Col.ImpRespBW *
        numpy.cos(numpy.deg2rad(sicd.SCPCOA.SlopeAng)))


#########################
# methods for calculating information density and rniirs

def get_information_density(
        bandwidth_area: Union[float, numpy.ndarray],
        signal: Union[float, numpy.ndarray],
        noise: Union[float, numpy.ndarray]) -> Union[float, numpy.ndarray]:
    """
    Calculate the information density from bandwidth area and signal/noise estimates.

    Parameters
    ----------
    bandwidth_area : float|numpy.ndarray
    signal : float|numpy.ndarray
    noise : float|numpy.ndarray

    Returns
    -------
    float|numpy.ndarray
    """

    return bandwidth_area * numpy.log2(1 + signal/noise)


def get_rniirs(
        information_density: Union[float, numpy.ndarray]) -> Union[float, numpy.ndarray]:
    r"""
    Calculate an RNIIRS estimate from the information density or
    Shannon-Hartley channel capacity.

    This mapping has been empirically determined by fitting Shannon-Hartley channel
    capacity to RNIIRS for some sample images.

    The basic model is given by
    :math:`rniirs = a_0 + a_1*\log_2(information\_density)`

    To maintain positivity of the estimated rniirs, this transitions to a linear
    model :math:`rniirs = slope*information\_density` with slope given by
    :math:`slope = a_1/(iim\_transition*\log(2))` below the transition point at
    :math:`transition = \exp(1 - \log(2)*a_0/a_1)`.

    Parameters
    ----------
    information_density : float|numpy.ndarray

    Returns
    -------
    float|numpy.ndarray
    """

    a = RNIIRS_FIT_PARAMETERS
    iim_transition = numpy.exp(1 - numpy.log(2) * a[0] / a[1])
    slope = a[1] / (iim_transition * numpy.log(2))

    if not isinstance(information_density, numpy.ndarray):
        information_density = numpy.array(information_density, dtype='float64')
    orig_ndim = information_density.ndim
    if orig_ndim == 0:
        information_density = numpy.reshape(information_density, (1, ))

    out = numpy.empty(information_density.shape, dtype='float64')
    mask = (information_density > iim_transition)
    mask_other = ~mask
    if numpy.any(mask):
        out[mask] = a[0] + a[1]*numpy.log2(information_density[mask])
    if numpy.any(mask_other):
        out[mask_other] = slope*information_density[mask_other]

    if orig_ndim == 0:
        return float(out[0])
    return out


def get_information_density_for_rniirs(
        rniirs: Union[float, numpy.ndarray]) -> Union[float, numpy.ndarray]:
    """
    The inverse of :func:`get_rniirs`, this determines the information density
    which yields the given RNIIRS.

    *New in version 1.2.35.*

    Parameters
    ----------
    rniirs : float|numpy.ndarray

    Returns
    -------
    float|numpy.ndarray
    """

    a = RNIIRS_FIT_PARAMETERS
    iim_transition = numpy.exp(1 - numpy.log(2) * a[0] / a[1])
    slope = a[1] / (iim_transition * numpy.log(2))
    rniirs_transition = slope*iim_transition

    if not isinstance(rniirs, numpy.ndarray):
        rniirs = numpy.array(rniirs, dtype='float64')
    orig_ndim = rniirs.ndim
    if orig_ndim == 0:
        rniirs = numpy.reshape(rniirs, (1, ))

    out = numpy.empty(rniirs.shape, dtype='float64')
    mask = (rniirs > rniirs_transition)
    mask_other = ~mask

    if numpy.any(mask):
        out[mask] = numpy.exp2((rniirs[mask] - a[0])/a[1])
    if numpy.any(mask_other):
        out[mask_other] = rniirs[mask_other]/slope

    if orig_ndim == 0:
        return float(out[0])
    return out


def snr_to_rniirs(
        bandwidth_area: Union[float, numpy.ndarray],
        signal: Union[float, numpy.ndarray],
        noise: Union[float, numpy.ndarray]) -> Tuple[Union[float, numpy.ndarray], Union[float, numpy.ndarray]]:
    """
    Calculate the information_density and RNIIRS estimate from bandwidth area and
    signal/noise estimates.

    It is assumed that geometric effects for signal and noise have been accounted for
    (i.e. use SigmaZeroSFPoly), and signal and noise have each been averaged to a
    single pixel value.

    This mapping has been empirically determined by fitting Shannon-Hartley channel
    capacity to RNIIRS for some sample images.

    Parameters
    ----------
    bandwidth_area : float
    signal : float
    noise : float

    Returns
    -------
    information_density : float
    rniirs : float
    """

    information_density = get_information_density(bandwidth_area, signal, noise)
    rniirs = get_rniirs(information_density)
    return information_density, rniirs


def rgiqe(sicd: SICDType) -> Tuple[float, float]:
    """
    Calculate the information_density and (default) estimated RNIIRS for the
    given sicd.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    information_density : float
    rniirs : float
    """

    bandwidth_area = get_bandwidth_area(sicd)
    signal = get_default_signal_estimate(sicd)
    noise = get_sigma0_noise(sicd)
    return snr_to_rniirs(bandwidth_area, signal, noise)


def populate_rniirs_for_sicd(
        sicd: SICDType,
        signal: Optional[float] = None,
        noise: Optional[float] = None,
        override: bool = False) -> None:
    """
    This populates the value(s) for RNIIRS and information density in the SICD
    structure, according to the RGIQE. **This modifies the sicd structure in place.**

    Parameters
    ----------
    sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    signal : None|float
        The signal value, in sigma zero.
    noise : None|float
        The noise equivalent sigma zero value.
    override : bool
        Override the value, if present.
    """

    if sicd.CollectionInfo is None:
        logger.error(
            'CollectionInfo must not be None.\n\t'
            'Nothing to be done for calculating RNIIRS.')
        return

    if sicd.CollectionInfo.Parameters is not None and \
            sicd.CollectionInfo.Parameters.get('PREDICTED_RNIIRS', None) is not None:
        if override:
            logger.info('PREDICTED_RNIIRS already populated, and this value will be overridden.')
        else:
            logger.info('PREDICTED_RNIIRS already populated. Nothing to be done.')
            return

    if noise is None:
        try:
            noise = get_sigma0_noise(sicd)
        except Exception as e:
            logger.error(
                'Encountered an error estimating noise for RNIIRS.\n\t{}'.format(e))
            return

    if signal is None:
        signal = get_default_signal_estimate(sicd)

    try:
        bw_area = get_bandwidth_area(sicd)
    except Exception as e:
        logger.error(
            'Encountered an error estimating bandwidth area for RNIIRS\n\t{}'.format(e))
        return

    inf_density, rniirs = snr_to_rniirs(bw_area, signal, noise)
    logger.info(
        'Calculated INFORMATION_DENSITY = {0:0.5G},\n\t'
        'PREDICTED_RNIIRS = {1:0.5G}'.format(inf_density, rniirs))
    if sicd.CollectionInfo.Parameters is None:
        sicd.CollectionInfo.Parameters = {}  # initialize
    sicd.CollectionInfo.Parameters['INFORMATION_DENSITY'] = '{0:0.2G}'.format(inf_density)
    sicd.CollectionInfo.Parameters['PREDICTED_RNIIRS'] = '{0:0.1f}'.format(rniirs)


def get_bandwidth_noise_distribution(
        sicd: SICDType,
        alpha: Union[float, numpy.ndarray],
        desired_information_density: Optional[float] = None,
        desired_rniirs: Optional[float] = None
        ) -> Tuple[Union[Tuple[float, float], numpy.ndarray], Union[float, numpy.ndarray]]:
    r"""
    This function determines SICD degradation parameters (nominally symmetric in
    row/column subaperture degradation) to achieve the desired information density/rniirs.

    There is natural one parameter distribution of reducing bandwidth and adding noise
    to achieve the desired RNIIRS/information density, based on :math:`\alpha \in [0, 1]`.

    The nominal relation is

    .. math::

        desired\_information\_density = bandwidth\_area*(bw\_mult(\alpha))^2)\cdot\log_2\left(
        1 + signal/(noise*noise\_mult(\alpha))\right).

    For :math:`\alpha=0`, we add no additional noise (:math:`bw\_mult(0) = bw\_min,noise\_mult(0)=1`)
    and use purely sub-aperture degradation, achieved at

    .. math::

        desired\_information\_density = bandwidth\_area*(bw\_min)^2)\cdot\log_2(1 + snr).

    On the other end, at :math:`\alpha=1`, we have :math:`bw\_mult(1)=1, noise\_mult(1)=noise\_mult\_max`
    and derive the noise multiplier which fulfills the required information density.
    For intermediate :math:`0 < \alpha < 1`, we find

    .. math::

        bw_mult(\alpha) = bw\_min\cdot (1-\alpha)

    Then, we find :math:`noise\_multiplier(\alpha)` which fulfills the required
    information density.

    .. note::

        Choosing subaperture windows is fundamentally a discrete operation, and
        this carries over to the realities of choosing bandwidth multipliers. See
        :func:`get_bidirectional_bandwidth_multiplier_possibilities` for all
        feasible values for bandwidth multipliers and associated noise adjustment
        details.

    *Refined in version 1.2.35.*

    Parameters
    ----------
    sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    alpha : float|numpy.ndarray
    desired_information_density : None|float
    desired_rniirs : None|float

    Returns
    -------
    bandwidth_multiplier : (float, float)|numpy.ndarray
        The `(row, column)` bandwidth multiplier including the discrete nature
        of this aperture, so the two may not be precisely equal.
    noise_multiplier : float|numpy.ndarray
        The noise multiplier, indicating how much noise to add before the
        subaperture processing.
    """

    # validate the desired information density/rniirs
    if (desired_information_density is None and desired_rniirs is None) or \
            (desired_information_density is not None and desired_rniirs is not None):
        raise ValueError('Exactly one of desired_information_density and desired_rniirs must be provided')

    if not isinstance(alpha, numpy.ndarray):
        alpha = numpy.array(alpha, dtype='float64')
    orig_ndim = alpha.ndim
    if orig_ndim == 0:
        alpha = numpy.reshape(alpha, (1, ))

    if not numpy.all((alpha >= 0) & (alpha <= 1)):
        raise ValueError('values for alpha must be in the interval [0, 1]')

    # get the current information density
    bandwidth_area = get_bandwidth_area(sicd)
    signal = get_default_signal_estimate(sicd)  # NB: this is just 1 or 0.25, no scaling issues
    current_nesz = get_sigma0_noise(sicd)
    snr = signal/current_nesz
    current_inf_density = get_information_density(bandwidth_area, signal, current_nesz)

    if desired_information_density is not None:
        desired_information_density = float(desired_information_density)
    elif desired_rniirs is not None:
        desired_information_density = get_information_density_for_rniirs(float(desired_rniirs))

    if desired_information_density > current_inf_density:
        raise ValueError(
            'The desired information density is {},\n\t'
            'but the current deweighted information density is {}'.format(
                desired_information_density, current_inf_density))

    aperture_size, bw_multiplier = get_bidirectional_bandwidth_multiplier_possibilities(sicd)

    # construct the whole list of bandwidth areas and resulting noises after
    # subaperture degrading and deweighting
    bw_areas = bandwidth_area*numpy.multiply.reduce(bw_multiplier, 1)
    inf_densities = get_information_density(bw_areas, signal, current_nesz)

    if desired_information_density < inf_densities[-1]:
        raise ValueError(
            'The desired information density is {},\n\t'
            'but the minimum possible with pure subaperture degradation is {}'.format(
                desired_information_density, inf_densities[-1]))

    best_index = numpy.argmin((desired_information_density - inf_densities)**2)

    indices = numpy.asarray(best_index - alpha*best_index, dtype=numpy.int32)
    indices = numpy.clip(indices, 0, best_index)

    this_bw_areas = bw_areas[indices]

    # NB: inf_dens = bw_area*log2(1 + snr/mult))
    #   snr/mult = 2^(inf_dens/bw_area) - 1
    #   mult = snr/(2^(inf_dens/bw_area) - 1))

    required_noise_multiplier = snr/(numpy.exp2(desired_information_density/this_bw_areas) - 1)
    required_noise_multiplier[required_noise_multiplier < 1] = 1

    bw_mult_out = numpy.empty(required_noise_multiplier.shape + (2, ), dtype='float64')
    bw_mult_out[:, 0] = bw_multiplier[indices, 0]
    bw_mult_out[:, 1] = bw_multiplier[indices, 1]

    if orig_ndim == 0:
        return (float(bw_mult_out[0, 0]), float(bw_mult_out[0, 1])), float(required_noise_multiplier[0])
    return bw_mult_out, required_noise_multiplier


#########################
# helpers for quality degradation function

def _get_uniform_weight_dicts(
        sicd: SICDType) -> Tuple[Optional[Dict], Optional[Dict]]:
    """
    Gets the dictionaries denoting uniform weighting.

    Parameters
    ----------
    sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    row_weighting : None|dict
    column_weighting : None|dict
    """

    row_weighting = None if is_uniform_weight(sicd, 0) else \
        {'WindowName': 'UNIFORM', 'WgtFunct': numpy.ones((32,), dtype='float64')}
    column_weighting = None if is_uniform_weight(sicd, 1) else \
        {'WindowName': 'UNIFORM', 'WgtFunct': numpy.ones((32,), dtype='float64')}
    return row_weighting, column_weighting


def _validate_reader(
        reader: Union[str, SICDTypeReader],
        index: int) -> Tuple[SICDTypeReader, int]:
    """
    Validate the method input:

    Parameters
    ----------
    reader : str|SICDTypeReader
    index : int
        The reader index.

    Returns
    -------
    reader: SICDTypeReader
    index: int
    """

    if isinstance(reader, str):
        reader = open_complex(reader)

    if not isinstance(reader, SICDTypeReader):
        raise TypeError('reader input must be a path to a complex file, or a sicd type reader instance')
    index = int(index)
    if not (0 <= index < reader.image_count):
        raise ValueError('index must be between 0 and {}, got {}'.format(reader.image_count, index))
    return reader, index


def _map_desired_resolution_to_aperture(
        current_imp_resp_bw: float,
        sample_size: float,
        direction: str,
        direction_size: int,
        desired_resolution: Optional[float] = None,
        desired_bandwidth: Optional[float] = None,
        broadening_factor: Optional[float] = None) -> Tuple[Optional[Tuple[int, int]], float]:
    """
    Determine the appropriate symmetric subaperture range to achieve the desired
    bandwidth or resolution, assuming the given broadening factor.

    Parameters
    ----------
    current_imp_resp_bw : float
    sample_size : float
    direction : str
    direction_size : int
        The size of the array along the given direction.
    desired_resolution : None|float
        The desired ImpRespWid (Row, Col) tuple, which will be mapped to ImpRespBW
        assuming uniform weighting. Exactly one of `desired_resolution` and
        `desired_bandwidth` must be provided.
    desired_bandwidth : None|float
        The desired ImpRespBW. Exactly one of `desired_resolution`
        and `desired_bandwidth` must be provided.
    broadening_factor : None|float
        The only applies if `desired_resolution` is provided. If not provided,
        then UNIFORM weighting will be assumed.

    Returns
    -------
    indices: None|Tuple[int, int]
    bw_factor : float
    """

    if desired_resolution is None and desired_bandwidth is None:
        raise ValueError('One of desire_resolution or desired_bandwidth must be supplied.')

    if desired_resolution is not None:
        if broadening_factor is None:
            broadening_factor = get_hamming_broadening_factor(1.0)
        else:
            broadening_factor = float(broadening_factor)
        use_resolution = float(desired_resolution)

        use_bandwidth = broadening_factor/use_resolution
    else:
        use_bandwidth = float(desired_bandwidth)

    if use_bandwidth > current_imp_resp_bw:
        if desired_resolution is not None:
            raise ValueError(
                'After mapping from Desired {} ImpRespWid considering uniform weighting,\n\t'
                'the equivalent desired ImpRespBW is {},\n\t'
                'but the current ImpRespBW is {}'.format(direction, use_bandwidth, current_imp_resp_bw))
        else:
            raise ValueError(
                'Desired {} ImpRespBW is given as {},\n\t'
                'but the current ImpRespBW is {}'.format(direction, use_bandwidth, current_imp_resp_bw))
    elif use_bandwidth == current_imp_resp_bw:
        return None, 1.0
    else:
        oversample = max(1., 1./(sample_size*use_bandwidth))
        ap_size = round(direction_size/oversample)
        start_ind = int(numpy.floor(0.5*(direction_size - ap_size)))
        return (start_ind, start_ind+ap_size), use_bandwidth/current_imp_resp_bw


def _map_bandwidth_parameters(
        sicd: SICDType,
        desired_resolution: Optional[Tuple[float, float]] = None,
        desired_bandwidth: Optional[Tuple[float, float]] = None
        ) -> Tuple[Tuple[int, int], float, Tuple[int, int], float]:
    """
    Helper function to map desired resolution or bandwidth to the suitable (centered)
    aperture.

    Parameters
    ----------
    sicd : SICDType
    desired_resolution : None|Tuple[float, float]
    desired_bandwidth : None|Tuple[float, float]

    Returns
    -------
    row_aperture : Tuple[int, int]
    row_bw_factor : float
    column_aperture : Tuple[int, int]
    column_bw_factor : float
    """

    if desired_resolution is not None:
        # get the broadening factor for uniform weighting
        broadening_factor = get_hamming_broadening_factor(1.0)
        row_aperture, row_bw_factor = _map_desired_resolution_to_aperture(
            sicd.Grid.Row.ImpRespBW, sicd.Grid.Row.SS, 'Row', sicd.ImageData.NumRows,
            desired_resolution=desired_resolution[0], broadening_factor=broadening_factor)
        column_aperture, column_bw_factor = _map_desired_resolution_to_aperture(
            sicd.Grid.Col.ImpRespBW, sicd.Grid.Col.SS, 'Col', sicd.ImageData.NumCols,
            desired_resolution=desired_resolution[1], broadening_factor=broadening_factor)
    elif desired_bandwidth is not None:
        row_aperture, row_bw_factor = _map_desired_resolution_to_aperture(
            sicd.Grid.Row.ImpRespBW, sicd.Grid.Row.SS, 'Row', sicd.ImageData.NumRows,
            desired_bandwidth=desired_bandwidth[0])
        column_aperture, column_bw_factor = _map_desired_resolution_to_aperture(
            sicd.Grid.Col.ImpRespBW, sicd.Grid.Col.SS, 'Col', sicd.ImageData.NumCols,
            desired_bandwidth=desired_bandwidth[1])
    else:
        row_aperture, row_bw_factor = None, 1
        column_aperture, column_bw_factor = None, 1

    return row_aperture, row_bw_factor, column_aperture, column_bw_factor


def get_dimension_bandwidth_multiplier_possibilities(
        sicd: SICDType,
        dimension: int) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Gets the bandwidth possibilities for all centered subapertures along the given
    dimension.

    *Introduced in 1.2.35*

    Parameters
    ----------
    sicd : SICDType
    dimension : int
        One of `{0, 1}`.

    Returns
    -------
    aperture_size : numpy.ndarray
        Of shape `(N, )`
    bandwidth_multiplier : numpy.ndarray
        Of shape `(N, )`
    """

    if dimension == 0:
        ap_size = round(sicd.ImageData.NumRows / sicd.Grid.Row.get_oversample_rate())
    else:
        ap_size = round(sicd.ImageData.NumCols / sicd.Grid.Col.get_oversample_rate())

    aperture_size = numpy.arange(ap_size, 0, -1, dtype='int32')
    bandwidth_multiplier = aperture_size/float(ap_size)

    return aperture_size, bandwidth_multiplier


def get_bidirectional_bandwidth_multiplier_possibilities(
        sicd: SICDType) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Gets the bandwidth possibilities for all centered subapertures shrinking
    along both dimensions symmetrically.

    *Introduced in 1.2.35*

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    aperture_size : numpy.ndarray
        An array of shape `(N, 2)` for row/column separately.
    bandwidth_multiplier : numpy.ndarray
        An array of shape `(N, 2)` for row/column separately.
    """

    row_aperture_size, row_bw_multiplier = get_dimension_bandwidth_multiplier_possibilities(sicd, 0)
    col_aperture_size, col_bw_multiplier = get_dimension_bandwidth_multiplier_possibilities(sicd, 1)

    the_size = max(row_aperture_size.size, col_aperture_size.size)
    aperture_size = numpy.empty((the_size, 2), dtype='int32')
    bandwidth_multiplier = numpy.empty((the_size, 2), dtype='float64')

    row_indexing = numpy.asarray(
        numpy.ceil(float(row_aperture_size.size - 1)*numpy.arange(the_size)/float(the_size - 1)), dtype=numpy.int32)
    col_indexing = numpy.asarray(
        numpy.ceil(float(col_aperture_size.size - 1)*numpy.arange(the_size)/float(the_size - 1)), dtype=numpy.int32)

    aperture_size[:, 0] = row_aperture_size[row_indexing]
    aperture_size[:, 1] = col_aperture_size[col_indexing]

    bandwidth_multiplier[:, 0] = row_bw_multiplier[row_indexing]
    bandwidth_multiplier[:, 1] = col_bw_multiplier[col_indexing]

    return aperture_size, bandwidth_multiplier


#########################
# SICD quality degradation functions

def quality_degrade(
        reader: Union[str, SICDTypeReader],
        index: int = 0,
        output_file: Optional[str] = None,
        desired_resolution: Optional[Tuple[float, float]] = None,
        desired_bandwidth: Optional[Tuple[float, float]] = None,
        desired_nesz: Optional[float] = None,
        **kwargs) -> Optional[FlatSICDReader]:
    r"""
    Create a degraded quality SICD based on the desired resolution (impulse response width)
    or bandwidth (impulse response bandwidth), and the desired Noise Equivalent
    Sigma Zero value. The produced SICD will have **uniform weighting**.

    No more than one of `desired_resolution` and `desired_bandwidth` can be provided.
    If None of `desired_resolution`, `desired_bandwidth`, or `desired_nesz` are provided,
    then the SICD will be re-weighted with uniform weighting - even this will
    change the noise and RNIIRS values slightly.

    .. warning::

        Unless `desired_nesz=None`, this will fail for a SICD which is not fully
        Radiometrically calibrated with `'ABSOLUTE'` noise type.


    Parameters
    ----------
    reader : str|SICDTypeReader
    index : int
        The reader index to be used.
    output_file : None|str
        If `None`, an in-memory SICD reader instance will be returned. Otherwise,
        this is the path for the produced output SICD file.
    desired_resolution : None|tuple
        The desired ImpRespWid (Row, Col) tuple, which will be mapped to ImpRespBW
        assuming uniform weighting. You cannot provide both `desired_resolution`
        and `desired_bandwidth`.
    desired_bandwidth : None|tuple
        The desired ImpRespBW (Row, Col) tuple. You cannot provide both
        `desired_resolution` and `desired_bandwidth`.
    desired_nesz : None|float
        The desired Noise Equivalent Sigma Zero value in power units, this is after
        modifications which change the noise due to sub-aperture degradation and/or
        de-weighting.
    kwargs
        Keyword arguments passed through to :func:`sarpy.processing.sicd.normalize_sicd.sicd_degrade_reweight`

    Returns
    -------
    None|FlatSICDReader
        No return if `output_file` is provided, otherwise the returns the in-memory
        reader object.
    """

    reader, index = _validate_reader(reader, index)
    if desired_resolution is not None and desired_bandwidth is not None:
        raise ValueError('Both desired_resolution and desired_bandwidth cannot be supplied.')

    sicd = reader.get_sicds_as_tuple()[index]

    if desired_nesz is None:
        add_noise = None
    else:
        current_nesz = get_sigma0_noise(sicd)
        add_noise_factor = (desired_nesz - current_nesz)/current_nesz
        if abs(add_noise_factor) < 1e-5:
            add_noise = None
        elif add_noise_factor < 0:
            raise ValueError(
                'The current nesz value is {},\n\t'
                'the desired nesz value of {} cannot be achieved.'.format(current_nesz, desired_nesz))
        else:
            add_noise = numpy.exp(numpy.log(10)*0.1*sicd.Radiometric.NoiseLevel.NoisePoly[0, 0])*add_noise_factor

    row_aperture, row_bw_factor, column_aperture, column_bw_factor = _map_bandwidth_parameters(
        sicd, desired_resolution=desired_resolution, desired_bandwidth=desired_bandwidth)

    row_weighting, column_weighting = _get_uniform_weight_dicts(sicd)
    return sicd_degrade_reweight(
        reader, output_file=output_file, index=index,
        row_aperture=row_aperture, row_weighting=row_weighting,
        column_aperture=column_aperture, column_weighting=column_weighting,
        add_noise=add_noise, **kwargs)


def quality_degrade_resolution(
        reader: Union[str, SICDTypeReader],
        index: int = 0,
        output_file: Optional[str] = None,
        desired_resolution: Optional[Tuple[float, float]] = None,
        desired_bandwidth: Optional[Tuple[float, float]] = None,
        **kwargs) -> Optional[FlatSICDReader]:
    """
    Create a degraded quality SICD based on INCREASING the impulse response width
    to the desired resolution or DECREASING the impulse response bandwidth to the
    desired bandwidth.

    The produced SICD will have uniform weighting.

    Parameters
    ----------
    reader : str|SICDTypeReader
    index : int
        The reader index to be used.
    output_file : None|str
        If `None`, an in-memory SICD reader instance will be returned. Otherwise,
        this is the path for the produced output SICD file.
    desired_resolution : None|tuple
        The desired ImpRespWid (Row, Col) tuple, which will be mapped to ImpRespBW
        assuming uniform weighting. Exactly one of `desired_resolution` and
        `desired_bandwidth` must be provided.
    desired_bandwidth : None|tuple
        The desired ImpRespBW (Row, Col) tuple. Exactly one of `desired_resolution`
        and `desired_bandwidth` must be provided.
    kwargs
        Keyword arguments passed through to :func:`sarpy.processing.sicd.normalize_sicd.sicd_degrade_reweight`

    Returns
    -------
    None|FlatSICDReader
        No return if `output_file` is provided, otherwise the returns the in-memory
        reader object.
    """

    return quality_degrade(
        reader, index=index, output_file=output_file,
        desired_resolution=desired_resolution, desired_bandwidth=desired_bandwidth,
        **kwargs)


def quality_degrade_noise(
        reader: Union[str, SICDTypeReader],
        index: int = 0,
        output_file: Optional[str] = None,
        desired_nesz: Optional[float] = None,
        **kwargs) -> Optional[FlatSICDReader]:
    """
    Create a degraded quality SICD based on INCREASING the noise to the desired
    Noise Equivalent Sigma Zero value. The produced SICD will have uniform weighting.

    .. warning::

        This will fail for a SICD which is not fully Radiometrically calibrated,
        with ABSOLUTE noise type.

    Parameters
    ----------
    reader : str|SICDTypeReader
    index : int
        The reader index to be used.
    output_file : None|str
        If `None`, an in-memory SICD reader instance will be returned. Otherwise,
        this is the path for the produced output SICD file.
    desired_nesz : None|float
        The desired noise equivalent sigma zero value.
    kwargs
        Keyword arguments passed through to :func:`sarpy.processing.sicd.normalize_sicd.sicd_degrade_reweight`

    Returns
    -------
    None|FlatSICDReader
        No return if `output_file` is provided, otherwise the returns the in-memory
        reader object.
    """

    return quality_degrade(reader, index=index, output_file=output_file, desired_nesz=desired_nesz, **kwargs)


def quality_degrade_rniirs(
        reader: Union[str, SICDTypeReader],
        index: int = 0,
        output_file: Optional[str] = None,
        desired_rniirs: Optional[float] = None,
        alpha: float = 0,
        **kwargs) -> Optional[FlatSICDReader]:
    r"""
    Create a degraded quality SICD based on the desired estimated RNIIRS value.
    The produced SICD will have uniform weighting.

    The sicd degradation will be performed as follows:

    - The current information density/current rniirs **with respect to uniform weighting**
      will be found.

    - The information density required to produce the desired rniirs will be found.

    - This will be used, along with the :math:`\alpha` value, will be used in
      :func:`get_bandwidth_noise_distribution` to determine the best feasible
      multipliers for bandwidth and noise values.

    - These desired bandwidth and noise values will then be used in conjunction
      with :func:`sicd_degrade_reweight`.

    .. warning::

        This will fail for a SICD which is not fully Radiometrically calibrated,
        with `'ABSOLUTE'` noise type.

    Parameters
    ----------
    reader : str|SICDTypeReader
    index : int
        The reader index to be used.
    output_file : None|str
        If `None`, an in-memory SICD reader instance will be returned. Otherwise,
        this is the path for the produced output SICD file.
    desired_rniirs : None|float
        The desired rniirs value, according to the RGIQE methodology.
    alpha : float
        This must be a number in the interval [0, 1] defining the (geometric)
        distribution of variability between required influence from increasing
        noise and require influence of decreasing bandwidth.
    kwargs
        Keyword arguments passed through to :func:`sarpy.processing.sicd.normalize_sicd.sicd_degrade_reweight`

    Returns
    -------
    None|FlatSICDReader
        No return if `output_file` is provided, otherwise the returns the in-memory
        reader object.
    """

    if desired_rniirs is None:
        return quality_degrade(reader, index=index, output_file=output_file, **kwargs)

    reader, index = _validate_reader(reader, index)
    sicd = reader.get_sicds_as_tuple()[index]
    current_noise = numpy.exp(numpy.log(10)*0.1*sicd.Radiometric.NoiseLevel.NoisePoly[0, 0])

    bandwidth_multiplier, noise_multiplier = get_bandwidth_noise_distribution(
        sicd, alpha, desired_rniirs=desired_rniirs)

    desired_bandwidth = (
        sicd.Grid.Row.ImpRespBW*bandwidth_multiplier[0],
        sicd.Grid.Col.ImpRespBW*bandwidth_multiplier[1])
    add_noise = (noise_multiplier - 1)*current_noise

    if alpha == 0 or add_noise <= 0:
        add_noise = None
    row_aperture, row_bw_factor, column_aperture, column_bw_factor = _map_bandwidth_parameters(
        sicd, desired_bandwidth=desired_bandwidth)
    row_weighting, column_weighting = _get_uniform_weight_dicts(sicd)

    return sicd_degrade_reweight(
        reader, output_file=output_file, index=index,
        row_aperture=row_aperture, row_weighting=row_weighting,
        column_aperture=column_aperture, column_weighting=column_weighting,
        add_noise=add_noise, **kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class AIMIDBType(TREElement):
    def __init__(self, value):
        super(AIMIDBType, self).__init__()
        self.add_field('ACQUISITION_DATE', 's', 14, value)
        self.add_field('MISSION_NO', 's', 4, value)
        self.add_field('MISSION_IDENTIFICATION', 's', 10, value)
        self.add_field('FLIGHT_NO', 's', 2, value)
        self.add_field('OP_NUM', 's', 3, value)
        self.add_field('CURRENT_SEGMENT', 's', 2, value)
        self.add_field('REPRO_NUM', 's', 2, value)
        self.add_field('REPLAY', 's', 3, value)
        self.add_field('RESERVED_001', 's', 1, value)
        self.add_field('START_TILE_COLUMN', 's', 3, value)
        self.add_field('START_TILE_ROW', 's', 5, value)
        self.add_field('END_SEGMENT', 's', 2, value)
        self.add_field('END_TILE_COLUMN', 's', 3, value)
        self.add_field('END_TILE_ROW', 's', 5, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('RESERVED002', 's', 4, value)
        self.add_field('LOCATION', 's', 11, value)
        self.add_field('RESERVED003', 's', 13, value)


class AIMIDB(TREExtension):
    _tag_value = 'AIMIDB'
    _data_type = AIMIDBType



"""
The ProductProcessingType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from collections import OrderedDict
from xml.etree import ElementTree
from typing import List

from sarpy.io.xml.base import Serializable, ParametersCollection, \
    find_children, find_first_child, get_node_value, create_text_node
from sarpy.io.xml.descriptors import ParametersDescriptor, StringDescriptor

from .base import DEFAULT_STRICT


class ProcessingModuleType(Serializable):
    """
    Flexibly structured processing module definition to keep track of the name and any parameters associated
    with the algorithms used to produce the SIDD.
    """

    _fields = ('ModuleName', 'name', 'ModuleParameters')
    _required = ('ModuleName', 'name', 'ModuleParameters')
    _set_as_attribute = ('name', )
    _collections_tags = {
        'ModuleParameters': {'array': False, 'child_tag': 'ModuleParameter'}}
    # Descriptor
    ModuleName = StringDescriptor(
        'ModuleName', _required, strict=DEFAULT_STRICT,
        docstring='The module name.')  # type: str
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='The module identifier.')  # type: str
    ModuleParameters = ParametersDescriptor(
        'ModuleParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Free form parameters.')  # type: ParametersCollection

    def __init__(self, ModuleName=None, name=None, ModuleParameters=None, ProcessingModules=None, **kwargs):
        """

        Parameters
        ----------
        ModuleName : str
        name : str
        ModuleParameters : None|ParametersCollection|dict
        ProcessingModules : None|List[ProcessingModuleType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ModuleName = ModuleName
        self.name = name
        self.ModuleParameters = ModuleParameters

        self._ProcessingModules = []
        if ProcessingModules is None:
            pass
        elif isinstance(ProcessingModules, ProcessingModuleType):
            self.addProcessingModule(ProcessingModules)
        elif isinstance(ProcessingModules, (list, tuple)):
            for el in ProcessingModules:
                self.addProcessingModule(el)
        else:
            raise ValueError('ProcessingModules got unexpected type {}'.format(type(ProcessingModules)))
        super(ProcessingModuleType, self).__init__(**kwargs)

    @property
    def ProcessingModules(self):
        """List[ProcessingModuleType]: list of ProcessingModules."""
        return self._ProcessingModules

    def getProcessingModule(self, key):
        """
        Get ProcessingModule(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[ProcessingModuleType]
        """

        return [entry for entry in self._ProcessingModules if entry.name == key]

    def addProcessingModule(self, value):
        """
        Add the ProcessingModule to the list.

        Parameters
        ----------
        value : ProcessingModuleType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            pm_key = self._child_xml_ns_key.get('ProcessingModules', self._xml_ns_key)
            value = ProcessingModuleType.from_node(value, self._xml_ns, ns_key=pm_key)
        elif isinstance(value, dict):
            value = ProcessingModuleType.from_dict(value)

        if isinstance(value, ProcessingModuleType):
            self._ProcessingModules.append(value)
        else:
            raise TypeError('Trying to set ProcessingModule with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        # parse the ModuleName
        mn_key = cls._child_xml_ns_key.get('ModuleName', ns_key)
        mn_node = find_first_child(node, 'ModuleName', xml_ns, mn_key)
        kwargs['ModuleName'] = get_node_value(mn_node)
        kwargs['name'] = mn_node.attrib.get('name', None)
        # parse the ProcessingModule children
        pm_key = cls._child_xml_ns_key.get('ProcessingModules', ns_key)
        kwargs['ProcessingModules'] = find_children(node, 'ProcessingModule', xml_ns, pm_key)
        return super(ProcessingModuleType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        exclude = exclude + ('ModuleName', 'name', 'ProcessingModules', 'ModuleParameters')
        node = super(ProcessingModuleType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # add the ModuleName and name children
        if self.ModuleName is not None:
            mn_key = self._child_xml_ns_key.get('ModuleName', ns_key)
            mn_tag = '{}:ModuleName'.format(mn_key) if mn_key is not None and mn_key != 'default' else 'ModuleName'
            mn_node = create_text_node(doc, mn_tag, self.ModuleName, parent=node)
            if self.name is not None:
                mn_node.attrib['name'] = self.name
        # add ModuleParameters
        if self.ModuleParameters:
            self.ModuleParameters.to_node(doc, ns_key=ns_key, parent=node, strict=strict)
        # add the ProcessingModule children
        pm_key = self._child_xml_ns_key.get('ProcessingModules', ns_key)
        for entry in self._ProcessingModules:
            entry.to_node(doc, tag, ns_key=pm_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(ProcessingModuleType, self).to_dict(check_validity=check_validity, strict=strict, exclude=exclude)
        if len(self.ProcessingModules) > 0:
            out['ProcessingModules'] = [
                entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._ProcessingModules]
        return out


class ProductProcessingType(Serializable):
    """
    Computed metadata regarding one or more of the input collections and final product.
    """

    _fields = ()
    _required = ()

    def __init__(self, ProcessingModules=None, **kwargs):
        """

        Parameters
        ----------
        ProcessingModules : None|List[ProcessingModuleType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        self._ProcessingModules = []
        if ProcessingModules is None:
            pass
        elif isinstance(ProcessingModules, ProcessingModuleType):
            self.addProcessingModule(ProcessingModules)
        elif isinstance(ProcessingModules, (list, tuple)):
            for el in ProcessingModules:
                self.addProcessingModule(el)
        else:
            raise ValueError('ProcessingModules got unexpected type {}'.format(type(ProcessingModules)))
        super(ProductProcessingType, self).__init__(**kwargs)

    @property
    def ProcessingModules(self):
        """List[ProcessingModuleType]: list of ProcessingModules."""
        return self._ProcessingModules

    def getProcessingModule(self, key):
        """
        Get ProcessingModule(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[ProcessingModuleType]
        """

        return [entry for entry in self._ProcessingModules if entry.name == key]

    def addProcessingModule(self, value):
        """
        Add the ProcessingModule to the list.

        Parameters
        ----------
        value : ProcessingModuleType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            pm_key = self._child_xml_ns_key.get('ProcessingModules', self._xml_ns_key)
            value = ProcessingModuleType.from_node(value, self._xml_ns, ns_key=pm_key)
        elif isinstance(value, dict):
            value = ProcessingModuleType.from_dict(value)

        if isinstance(value, ProcessingModuleType):
            self._ProcessingModules.append(value)
        else:
            raise TypeError('Trying to set ProcessingModule with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()

        pm_key = cls._child_xml_ns_key.get('ProcessingModules', ns_key)
        kwargs['ProcessingModules'] = find_children(node, 'ProcessingModule', xml_ns, pm_key)
        return super(ProductProcessingType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(ProductProcessingType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the ProcessingModule children
        pm_key = self._child_xml_ns_key.get('ProcessingModules', ns_key)
        for entry in self._ProcessingModules:
            entry.to_node(doc, 'ProcessingModule', ns_key=pm_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(ProductProcessingType, self).to_dict(check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        if len(self.ProcessingModules) > 0:
            out['ProcessingModules'] = [
                entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._ProcessingModules]
        return out




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PATCHA_74Type(TREElement):
    def __init__(self, value):
        super(PATCHA_74Type, self).__init__()
        self.add_field('PAT_NO', 's', 4, value)
        self.add_field('LAST_PAT_FLAG', 's', 1, value)
        self.add_field('LNSTRT', 's', 7, value)
        self.add_field('LNSTOP', 's', 7, value)
        self.add_field('AZL', 's', 5, value)
        self.add_field('NVL', 's', 5, value)
        self.add_field('FVL', 's', 3, value)
        self.add_field('NPIXEL', 's', 5, value)
        self.add_field('FVPIX', 's', 5, value)
        self.add_field('FRAME', 's', 3, value)
        self.add_field('GMT', 's', 8, value)
        self.add_field('SHEAD', 's', 7, value)
        self.add_field('GSWEEP', 's', 6, value)
        self.add_field('SHEAR', 's', 8, value)


class PATCHA_74(TREExtension):
    _tag_value = 'PATCHA'
    _data_type = PATCHA_74Type


class PATCHA_115Type(TREElement):
    def __init__(self, value):
        super(PATCHA_115Type, self).__init__()
        self.add_field('PAT_NO', 's', 4, value)
        self.add_field('LAST_PAT_FLAG', 's', 1, value)
        self.add_field('LNSTRT', 's', 7, value)
        self.add_field('LNSTOP', 's', 7, value)
        self.add_field('AZL', 's', 5, value)
        self.add_field('NVL', 's', 5, value)
        self.add_field('FVL', 's', 3, value)
        self.add_field('NPIXEL', 's', 5, value)
        self.add_field('FVPIX', 's', 5, value)
        self.add_field('FRAME', 's', 3, value)
        self.add_field('UTC', 's', 8, value)
        self.add_field('SHEAD', 's', 7, value)
        self.add_field('GRAVITY', 's', 7, value)
        self.add_field('INS_V_NC', 's', 5, value)
        self.add_field('INS_V_EC', 's', 5, value)
        self.add_field('INS_V_DC', 's', 5, value)
        self.add_field('OFFLAT', 's', 8, value)
        self.add_field('OFFLONG', 's', 8, value)
        self.add_field('TRACK', 's', 3, value)
        self.add_field('GSWEEP', 's', 6, value)
        self.add_field('SHEAR', 's', 8, value)


class PATCHA_115(TREExtension):
    _tag_value = 'PATCHA'
    _data_type = PATCHA_115Type


class PATCHA(TREExtension):
    _tag_value = 'PATCHA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one PATCHA_74 or PATCHA_115')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        PATCHA_74|PATCHA_115
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 74:
            return PATCHA_74.from_bytes(value, start)
        elif lng == 115:
            return PATCHA_115.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 74 or 115. Got {}'.format(lng))



"""
The GeoInfo definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from collections import OrderedDict
from xml.etree import ElementTree
from typing import List, Dict

import numpy
from sarpy.io.xml.base import Serializable, ParametersCollection, \
    find_children, create_new_node
from sarpy.io.xml.descriptors import StringDescriptor, \
    ParametersDescriptor, SerializableListDescriptor
from sarpy.io.complex.sicd_elements.blocks import LatLonRestrictionType, LatLonArrayElementType

from .base import DEFAULT_STRICT

class LineType(Serializable):
    _fields = ('Endpoint', 'size')
    _required = ('Endpoint', 'size')

    def __init__(self, Endpoint=None, **kwargs):
        """

        Parameters
        ----------
        Endpoint : List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        kwargs
        """

        self._array = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Endpoint = Endpoint
        super(LineType, self).__init__(**kwargs)

    @property
    def size(self):
        """
        int: The size attribute
        """

        return 0 if self._array is None else self._array.size

    @property
    def Endpoint(self):
        """
        numpy.ndarray: The array of points.
        """

        return numpy.array([], dtype='object') if self._array is None else self._array

    @Endpoint.setter
    def Endpoint(self, value):
        if value is None:
            self._array = None
            return

        # LineType must have at least 2 Elements
        if len(value) < 2:
            raise ValueError(f'LineType must have at least 2 endpoints, got {len(value)}')

        if isinstance(value, numpy.ndarray):
            is_type = True
            for entry in value:
                is_type &= isinstance(entry, LatLonArrayElementType)
            if is_type:
                self._array = value
                return

        if isinstance(value, (numpy.ndarray, list, tuple)):
            use_value = []
            for i, entry in enumerate(value):
                if isinstance(entry, LatLonArrayElementType):
                    entry.index = i+1
                    use_value.append(entry)
                elif isinstance(entry, dict):
                    e_val = LatLonArrayElementType.from_dict(entry)
                    e_val.index = i+1
                    use_value.append(e_val)
                elif isinstance(entry, (numpy.ndarray, list, tuple)):
                    use_value.append(LatLonArrayElementType.from_array(entry, index=i+1))
                else:
                    raise TypeError('Got unexpected type for element of Endpoint array `{}`'.format(type(entry)))
            self._array = numpy.array(use_value, dtype='object')
        else:
            raise TypeError('Got unexpected type for Endpoint array `{}`'.format(type(value)))

    def __getitem__(self, item):
        return self._array.__getitem__(item)

    def __setitem__(self, key, value):
        self._array.__setitem__(key, value)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        """
        Parameters
        ----------
        node
        xml_ns : None|dict
        ns_key : None|str
        kwargs : dict

        Returns
        -------
        LineType
        """

        end_point_key = cls._child_xml_ns_key.get('Endpoint', ns_key)
        end_points = []
        for cnode in find_children(node, 'Endpoint', xml_ns, end_point_key):
            end_points.append(LatLonArrayElementType.from_node(cnode, xml_ns, ns_key=end_point_key))
        return cls(Endpoint=end_points)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        if parent is None:
            parent = doc.getroot()

        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)

        node.attrib['size'] = str(self.size)
        end_point_key = self._child_xml_ns_key.get('Endpoint', ns_key)

        for entry in self.Endpoint:
            entry.to_node(doc, 'Endpoint', ns_key=end_point_key, parent=node,
                          check_validity=check_validity, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        return OrderedDict([
            ('Endpoint', [entry.to_dict() for entry in self.Endpoint]),
            ('size', self.size)])


class PolygonType(Serializable):
    _fields = ('Vertex', 'size')
    _required = ('Vertex', 'size')

    def __init__(self, Vertex=None, **kwargs):
        """

        Parameters
        ----------
        Vertex : List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        kwargs
        """

        self._array = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Vertex = Vertex
        super(PolygonType, self).__init__(**kwargs)

    @property
    def size(self):
        """
        int: The size attribute
        """

        if self._array is None:
            return 0
        else:
            return self._array.size

    @property
    def Vertex(self):
        """
        numpy.ndarray: The array of points.
        """

        if self._array is None:
            return numpy.array((0,), dtype='object')
        else:
            return self._array

    @Vertex.setter
    def Vertex(self, value):
        if value is None:
            self._array = None
            return

        # PolygonType must have at least 3 Vertices
        if len(value) < 3:
            raise ValueError(f'PolygonType must have at least 3 vertices, got {len(value)}')

        if isinstance(value, numpy.ndarray):
            is_type = True
            for entry in value:
                is_type &= isinstance(entry, LatLonArrayElementType)
            if is_type:
                self._array = value
                return

        if isinstance(value, (numpy.ndarray, list, tuple)):
            use_value = []
            for i, entry in enumerate(value):
                if isinstance(entry, LatLonArrayElementType):
                    entry.index = i + 1
                    use_value.append(entry)
                elif isinstance(entry, dict):
                    e_val = LatLonArrayElementType.from_dict(entry)
                    e_val.index = i + 1
                    use_value.append(e_val)
                elif isinstance(entry, (numpy.ndarray, list, tuple)):
                    use_value.append(LatLonArrayElementType.from_array(entry, index=i + 1))
                else:
                    raise TypeError('Got unexpected type for element of Vertex array `{}`'.format(type(entry)))
            self._array = numpy.array(use_value, dtype='object')
        else:
            raise TypeError('Got unexpected type for Vertex array `{}`'.format(type(value)))

    def __getitem__(self, item):
        return self._array.__getitem__(item)

    def __setitem__(self, key, value):
        self._array.__setitem__(key, value)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        """
        Parameters
        ----------
        node
        xml_ns : None|dict
        ns_key : None|str
        kwargs : dict

        Returns
        -------
        PolygonType
        """

        vertex_key = cls._child_xml_ns_key.get('Vertex', ns_key)
        vertices = []
        for cnode in find_children(node, 'Vertex', xml_ns, vertex_key):
            vertices.append(LatLonArrayElementType.from_node(cnode, xml_ns, ns_key=vertex_key))
        return cls(Vertex=vertices)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        if parent is None:
            parent = doc.getroot()

        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)

        node.attrib['size'] = str(self.size)
        end_point_key = self._child_xml_ns_key.get('Vertex', ns_key)

        for entry in self.Vertex:
            entry.to_node(doc, 'Vertex', ns_key=end_point_key, parent=node,
                          check_validity=check_validity, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        return OrderedDict([
            ('Vertex', [entry.to_dict() for entry in self.Vertex]),
            ('size', self.size)])


class GeoInfoType(Serializable):
    """
    A geographic feature.
    """

    _fields = ('name', 'Descriptions', 'Point', 'Line', 'Polygon', 'GeoInfo')
    _required = ('name', )
    _set_as_attribute = ('name', )
    _collections_tags = {
        'Descriptions': {'array': False, 'child_tag': 'Desc'},
        'Point': {'array': False, 'child_tag': 'Point'},
        'Line': {'array': False, 'child_tag': 'Line'},
        'Polygon': {'array': False, 'child_tag': 'Polygon'},
        'GeoInfo': {'array': False, 'child_tag': 'GeoInfo'}
    }
    # descriptors
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='The name.')  # type: str
    Descriptions = ParametersDescriptor(
        'Descriptions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Descriptions of the geographic feature.')  # type: ParametersCollection
    Point = SerializableListDescriptor(
        'Point', LatLonRestrictionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic points with WGS-84 coordinates.'
    )  # type: List[LatLonRestrictionType]
    Line = SerializableListDescriptor(
        'Line', LineType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic lines (array) with WGS-84 coordinates.'
    )  # type: List[LineType]
    Polygon = SerializableListDescriptor(
        'Polygon', PolygonType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic polygons (array) with WGS-84 coordinates.'
    )  # type: List[PolygonType]

    def __init__(self, name=None, Descriptions=None, Point=None, Line=None,
                 Polygon=None, GeoInfo=None, **kwargs):
        """

        Parameters
        ----------
        name : str
        Descriptions : ParametersCollection|dict
        Point : List[LatLonRestrictionType]
        Line : List[LineType]
        Polygon : List[PolygonType]
        GeoInfo : Dict[GeoInfoType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.name = name
        self.Descriptions = Descriptions
        self.Point = Point
        self.Line = Line
        self.Polygon = Polygon

        self._GeoInfo = []
        if GeoInfo is None:
            pass
        elif isinstance(GeoInfo, GeoInfoType):
            self.addGeoInfo(GeoInfo)
        elif isinstance(GeoInfo, (list, tuple)):
            for el in GeoInfo:
                self.addGeoInfo(el)
        else:
            raise ValueError('GeoInfo got unexpected type {}'.format(type(GeoInfo)))
        super(GeoInfoType, self).__init__(**kwargs)

    @property
    def GeoInfo(self):
        """
        List[GeoInfoType]: list of GeoInfo objects.
        """

        return self._GeoInfo

    def getGeoInfo(self, key):
        """
        Get GeoInfo(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfo if entry.name == key]

    def addGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfo list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfo', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfo.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfo', ns_key)
        kwargs['GeoInfo'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(GeoInfoType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeoInfoType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity,
            strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self._GeoInfo is not None and len(self._GeoInfo) > 0:
            for entry in self._GeoInfo:
                entry.to_node(doc, tag, ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeoInfoType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self.GeoInfo is not None and len(self.GeoInfo) > 0:
            out['GeoInfo'] = [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._GeoInfo]
        return out




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class LINE_NUM_COEFF(TREElement):
    def __init__(self, value):
        super(LINE_NUM_COEFF, self).__init__()
        self.add_field('LINE_NUM_COEFF', 's', 12, value)


class LINE_DEN_COEFF(TREElement):
    def __init__(self, value):
        super(LINE_DEN_COEFF, self).__init__()
        self.add_field('LINE_DEN_COEFF', 's', 12, value)


class SAMP_NUM_COEFF(TREElement):
    def __init__(self, value):
        super(SAMP_NUM_COEFF, self).__init__()
        self.add_field('SAMP_NUM_COEFF', 's', 12, value)


class SAMP_DEN_COEFF(TREElement):
    def __init__(self, value):
        super(SAMP_DEN_COEFF, self).__init__()
        self.add_field('SAMP_DEN_COEFF', 's', 12, value)


class RPC00AType(TREElement):
    def __init__(self, value):
        super(RPC00AType, self).__init__()
        self.add_field('SUCCESS', 's', 1, value)
        self.add_field('ERR_BIAS', 's', 7, value)
        self.add_field('ERR_RAND', 's', 7, value)
        self.add_field('LINE_OFF', 's', 6, value)
        self.add_field('SAMP_OFF', 's', 5, value)
        self.add_field('LAT_OFF', 's', 8, value)
        self.add_field('LONG_OFF', 's', 9, value)
        self.add_field('HEIGHT_OFF', 's', 5, value)
        self.add_field('LINE_SCALE', 's', 6, value)
        self.add_field('SAMP_SCALE', 's', 5, value)
        self.add_field('LAT_SCALE', 's', 8, value)
        self.add_field('LONG_SCALE', 's', 9, value)
        self.add_field('HEIGHT_SCALE', 's', 5, value)
        self.add_loop('LINE_NUM_COEFFs', 20, LINE_NUM_COEFF, value)
        self.add_loop('LINE_DEN_COEFFs', 20, LINE_DEN_COEFF, value)
        self.add_loop('SAMP_NUM_COEFFs', 20, SAMP_NUM_COEFF, value)
        self.add_loop('SAMP_DEN_COEFFs', 20, SAMP_DEN_COEFF, value)


class RPC00A(TREExtension):
    _tag_value = 'RPC00A'
    _data_type = RPC00AType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)


class BNDPLBType(TREElement):
    def __init__(self, value):
        super(BNDPLBType, self).__init__()
        self.add_field('NUMPTS', 'd', 4, value)
        self.add_loop('PTs', self.NUMPTS, PT, value)


class BNDPLB(TREExtension):
    _tag_value = 'BNDPLB'
    _data_type = BNDPLBType
